r""" The ActiveGalacticNucleus module implements the AGN contribution by loading templates from Fritz et al., 2006.

Templated emission is divided into 3 components:

- Accretion disk around the central SMBH (accessible through the instance variable :code:`AGN().disk`)
- Scattered emission by the surrounding dusty torus (accessible through the instance variable :code:`AGN().scat`)
- Thermal dust emission associated to the dusty torus (accessible through the instance variable :code:`AGN().ther`)

The templates are computed accounting for the variation of 6 structural parameters:

- the covering angle of the torus :math:`\Gamma`, expressed in terms of half the aperture-angle with respect to the equatorial plane:

  .. math ::

    \Theta = 90^\circ - \dfrac{\Gamma}{2}

- the dust density distribution in spherical coordinates:

  .. math ::

    \rho \propto r^\beta e^{- \alpha|\cos\theta|}

  parameterized in terms of :math:`\alpha` and :math:`\beta`;
- the ratio :math:`R^\text{AGN}_\text{torus}` between the maximum to minimum radii of the dusty torus;
- the optical depth :math:`\tau^\text{AGN}_{9.7}` at :math:`9.7\ \mu m`;
- the viewing angle :math:`\Psi^\text{AGN}_\text{los}`, i.e. the angle between the rotation axis and the line-of-sight.
  (note that by assuming the unifed AGN model, with :math:`\Psi^\text{AGN}_\text{los} = 90^\circ` the geometry of 
  the object corresponds to a type 1 AGN, while :math:`\Psi^\text{AGN}_\text{los} = 0^\circ` to a type 2 AGN)
  
The above parameters in our implementation are controlled through the above parameters following the original naming schema used by the authors and summarized in the following table:

+------------------------------------+------------+-------------------------+----------------------------------------------------------------+
| Parameter                          | Key        | :math:`N_\text{values}` | Values                                                         |
+====================================+============+=========================+================================================================+
| :math:`\Theta`                     | :code:`ct` | 3                       | (:math:`20^\circ,\ 40^\circ,\ 60^\circ`)                       |
+------------------------------------+------------+-------------------------+----------------------------------------------------------------+ 
| :math:`\alpha`                     | :code:`al` | 4                       | (:math:`0,\ 2,\ 4,\ 6`)                                        |
+------------------------------------+------------+-------------------------+----------------------------------------------------------------+
| :math:`\beta`                      | :code:`be` | 5                       | (:math:`-1,\ -0.75,\ -0.5,\ -0.25,\ 0`)                        |
+------------------------------------+------------+-------------------------+----------------------------------------------------------------+
| :math:`R^\text{AGN}_\text{torus}`  | :code:`rm` | 5                       | (:math:`10,\ 30,\ 60,\ 100,\ 150`)                             |
+------------------------------------+------------+-------------------------+----------------------------------------------------------------+
| :math:`\tau^\text{AGN}_{9.7}`      | :code:`ta` | 8                       | (:math:`0.1,\ 0.3,\ 0.6,\ 1,\ 2,\ 3,\ 6,\ 10`)                 |
+------------------------------------+------------+-------------------------+----------------------------------------------------------------+
| :math:`\Psi^\text{AGN}_\text{los}` | :code:`ia` | 10                      | :math:`0^\circ` to :math:`90^\circ` with step :math:`10^\circ` |
+------------------------------------+------------+-------------------------+----------------------------------------------------------------+

Which results in 24000 available templates. The user can nevertheless set the parameters to whatever value and the template whose parameters are closer to the selection is chosen.
"""

# External imports
import numpy

# Internal imports
import galapy.internal.globs as GP_GBL
from galapy.internal.utils import find_nearest, trap_int, powerlaw_exp_cutoff
from galapy.internal.constants import Ang_to_keV
from galapy.internal.interp import lin_interp
from galapy.internal.data import DataFile

_template_tunables = ( 'ct', 'al', 'be', 'ta', 'rm', 'ia' )

_template_pars = {
    'ct' : numpy.array( [ 20, 40, 60 ], dtype = int ),
    'al' : numpy.array( [ 0.0, 2.0, 4.0, 6.0 ], dtype = float ),
    'be' : numpy.array( [ -1.00, -0.75, -0.50, -0.25, 0.00 ], dtype = float ),
    'ta' : numpy.array( [ 0.1, 0.3, 0.6, 1.0, 2.0, 3.0, 6.0, 10.0 ], dtype = float ),
    'rm' : numpy.array( [ 10, 30, 60, 100, 150 ], dtype = int ),
    'ia' : numpy.array( [ 1.e-3, 10.1, 20.1, 30.1, 40.1, 50.1, 60.1, 70.1, 80.1, 89.99 ], dtype = float ),
}

def find_template_par ( key, value ) :
    """ Given a key-value pair returns the parameter corresponding to `key`
    included in the template-library and closest to `value`
    """
    try :
        return _template_pars[ key ][ find_nearest( _template_pars[ key ], value ) ]
    except KeyError :
        raise AttributeError( f"Parameter '{key}' is not a valid template parameter." )

def agn_build_params ( fAGN, **kwargs ) :
    """ Standard function for building the parameters dictionary of class AGN()
    """
    out = {
        'fAGN' : fAGN,
        'template' : {
            'ct' : 40,
            'al' : 0.,
            'be' : -0.5,
            'ta' : 6.,
            'rm' : 60,
            'ia' : 0.001
        }
    }
    for k in set( out['template'].keys() ).intersection( kwargs.keys() ) :
        out['template'][k] = find_template_par( k, kwargs[k] )
        
    return out

#################################################################################
# AGN class
#################################################################################

class AGN () :
    r""" AGN component class
    This class implements the templated emission from an Active Galactic Nucleus
    within the galaxy.
    We use templates from Fritz et al., 2006 to model the AGN emission normalized
    at build time. The total emission is then rescaled to a user-provided total emission.
    
    The final emission can be expressed as
    
    .. math::
    
       L_\text{AGN}(\lambda) = \dfrac{f_\text{AGN}}{1-f_\text{AGN}}L_\text{AGN}^\text{norm}(\lambda)\cdot L_\text{ref}
    
    where 
    :math:`f_\text{AGN}` is the fraction of the reference emission, :math:`L_\text{ref}`, 
    radiated by the AGN and
    :math:`L_\text{AGN}^\text{norm}` is the normalized total templated emission.
    
    Parameters
    ----------
    lmin, lmax : float
      minimum and maximum values of the wavelength-domain. The original templates
      are computed within the 10-10^7 Angstrom interval. At build time this domain
      is extended with `pad` padding values before and after the limits of the 
      template's wavelength domain to match the requested limits.
    pad : integer
      number of padding values to match the requested wavelength domain
    fAGN : float
      fraction of the total reference emission
    do_Xray : bool
      If :code:`True` compute the X-ray template along with the closest template for
      the lower energy part of the spectrum (default is :code:`True`)

    Keyword Arguments
    -----------------
    ct : scalar
      the covering angle of the torus :math:`\Gamma`, expressed in terms of half the aperture-angle with respect to the equatorial plane
    al : scalar
      parameter regulating the dust density distribution in spherical coordinates, corresponds to :math:`\alpha`
    be : scalar
      parameter regulating the dust density distribution in spherical coordinates, corresponds to :math:`\beta`
    ta : scalar
      the optical depth at :math:`9.7\ \mu m`
    rm : scalar
      the ratio between the maximum to minimum radii of the dusty torus      
    ia : scalar
      the inclination angle between the rotation axis of the AGN torus and the line of sight.
      Note that by assuming the unifed AGN model, with :code:`ia = 90` the geometry of the object corresponds to a type 1 AGN, 
      while :code:`ia = 0` to a type 2 AGN.
    """
    
    def __init__ ( self, lmin, lmax, pad = 16, fAGN = 1.e-3, do_Xray = True, **kwargs ) :
        import galapy.internal.globs as GP_GBL
        import os

        # check inputs
        if not fAGN < 1.0 :
            raise RuntimeError( 'trying to set fAGN>=1.0 not allowed.' )
        
        # store the argument variables
        self.lmin, self.lmax = lmin, lmax
        self._pad = pad
        self.do_Xray = do_Xray
        
        # common name of all template files
        self._filebase = GP_GBL.AGN_FILE

        # build the parameter dictionary
        self.params = agn_build_params( fAGN, **kwargs)

        # load the template with physics nearest to the parameters value 
        self.load_template()

        # also compute the X-ray template if requested (default=True)
        if self.do_Xray :
            self.compute_X_template()

    def load_template ( self ) :
        """ Loads the template corresponding to the current parameters (mostly intended for internal usage).
        """

        # load template from closest file to the parameters chosen
        filename = self._filebase.format( *( self.params[ 'template' ][k]
                                             for k in _template_tunables  ) )
        self.ll, self.tot, self.ther, self.scat, self.disk = \
            numpy.array( [1.e+4,1.e-4,1.e-4,1.e-4,1.e-4] )[:,numpy.newaxis] * \
            numpy.genfromtxt( DataFile( filename, GP_GBL.AGN_DIR ).get_file(),
                              unpack=True )

        # Extend the wavelength domain by padding with zeros the emissions
        ltmp = numpy.pad( self.ll, self._pad, constant_values = 0. )
        ltmp[:self._pad+1]  = numpy.logspace( numpy.log10(self.lmin),
                                              numpy.log10(self.ll.min()),
                                              self._pad+1 )
        ltmp[-self._pad-1:] = numpy.logspace( numpy.log10(self.ll.max()),
                                              numpy.log10(self.lmax),
                                              self._pad+1 )
        self.ll   = ltmp
        self.ther = numpy.pad(self.ther, self._pad, constant_values = 0.)
        self.scat = numpy.pad(self.scat, self._pad, constant_values = 0.)
        self.disk = numpy.pad(self.disk, self._pad, constant_values = 0.)
        self.tot  = numpy.pad(self.tot,  self._pad, constant_values = 0.)

        # Compute constant normalization factor for the current template 
        self.Lnorm = 1. / trap_int( self.ll, self.tot )

        # Build a linear normalized interpolator
        self.f_norm_tot = lin_interp( numpy.ascontiguousarray( self.ll ),
                                      numpy.ascontiguousarray( self.tot * self.Lnorm ) )
        return;

    def compute_X_template ( self ) :
        """ Pre-computes the not-normalized and not-bolometric-corrected X-ray spectrum. 
        """

        # generate wavelength grid
        ll = numpy.logspace( numpy.log10( self.lmin ),
                             numpy.log10( self.lmax ),
                             256 )
        # convert wavelength to energy
        El = Ang_to_keV( ll )

        if El.max() < 2. :
            raise RuntimeError( "Cannot build the X-ray spectrum for "
                                "a wavelength grid starting at lambda > "
                                "6 Angstrom ~ 2 keV! "
                                "Set a smaller `lmin` value." )
        
        # find interval for hard-X normalization
        wE = ( 2. <= El ) & ( El <= 10. )

        # compute emission law normalization
        Lnorm = -1. / trap_int( El[wE], powerlaw_exp_cutoff( El[wE], gamma=1.8, Ecut=3.e+2 ) )

        # Compute the normalized power-law:
        # - high energy cut-off at 300 keV
        # - spectral index fixed to 1.8
        # - low energy cut-off at 50 Angstrom (~0.25 keV)
        ret = numpy.zeros_like( ll )
        wL = ( ll <= 5.e+1 ) 
        ret[wL] = powerlaw_exp_cutoff( El[wL], gamma=1.8, Ecut=3.e+2 ) * Lnorm

        # store normalized interpolator object for X-ray emission
        self.f_norm_X = lin_interp( numpy.ascontiguousarray( ll ),
                                    numpy.ascontiguousarray( ret ) )
        return;
        

    def X_bolometric_correction ( self, Lref ) :
        r""" Computes the bolometric correction from Duras et al., 2020 (Eq. 2):
        
        .. math ::
        
          \dfrac{L_\text{bol}^\text{AGN}}{L_\text{X}^\text{AGN}} \approx 10.96\;\biggl[1 + \biggl(\dfrac{\log L_\text{bol}^\text{AGN}/L_\odot}{11.93}\biggr)^17.79\biggr]

        Parameters
        ----------
        Lref : scalar
          Bolometric luminosity of the AGN in units of solar luminosities :math:`L_\odot`
        
        Returns
        -------
        : scalar
          Value of the bolometric correction.
        """
        return 10.96 * ( 1. + ( numpy.log10( Lref ) / 11.93 )**17.79 )
        
    def set_parameters ( self, fAGN = None, **kwargs ) :
        r""" Sets the internal parameters defining the emission coming from the AGN.
        All the parameters that are not passed to this function will be left to the
        value they already have.
        
        Parameters
        ----------
        fAGN : scalar
          sets the normalization of the templated emission. The parameters fAGN 
          provides the amount of energy radiated by the AGN in units of some other
          emission contributor, typically the energy radiated by dust in the same band.
        **kwargs 
          These are the parameters passed to function :code:`find_template_par`. Used
          to find the nearest AGN emission template in the Fritz et al. (2006) library.
        """

        if fAGN is not None :
            if fAGN < 1.0 :
                self.params.update( fAGN = fAGN )
            else :
                raise RuntimeError( 'trying to set fAGN>=1.0 not allowed.' )
        if len(kwargs) > 0 :
            for k in set( self.params['template'].keys() ).intersection( kwargs.keys() ) :
                self.params['template'][k] = find_template_par( k, kwargs[k] )
            self.load_template()
        return;

    def emission ( self, ll, Lref ) :
        """ Computes the emission coming from the AGN adding the X-ray part 
        if the internal variable do_Xray has been set to :code:`True` at build time
        
        Parameters
        ----------
        ll : sequence
          wavelength list or array or iterable
        Lref : float
          bolometric reference luminosity used to compute the bolometric correction
          of the high-energy part of the spectrum
        
        Returns
        -------
        : numpy-array
          Luminosity per unit wavelength in units of solar luminosities emitted by
          the AGN.
        """
        ll = numpy.ascontiguousarray( ll, dtype = numpy.float64 )
        fact = self.params['fAGN']/(1-self.params['fAGN'])
        if self.do_Xray :
            return fact * Lref * ( self.f_norm_tot( ll ) +
                                   self.f_norm_X( ll ) *
                                   self.X_bolometric_correction( Lref ) )
        return  fact * Lref * self.f_norm_tot( ll )
    
        




from abc import ABC, abstractmethod, abstractclassmethod

class GalaPyObject ( ABC ) :
    
    @abstractmethod
    def _dump_hdf5 ( self, outfile ) :
        pass

    @abstractclassmethod
    def _load_hdf5 ( cls, indict ) :
        pass
    
    def dump ( self, outfile, method = 'hdf5' ) :
        if method == 'hdf5' :
            self._dump_hdf5( outfile )
            return;
        if method == 'pickle' :
            import pickle
            with open( outfile, 'wb' ) as f :
                pickle.dump( self, f )
            return;
        return;

    @classmethod
    def load ( cls, infile, method = 'hdf5' ) :
        if method == 'hdf5' :
            from galapy.io import load_from_hdf5
            return _load_hdf5( load_from_hdf5( infile ) )
        if method == 'pickle' :
            import pickle
            with open( infile, 'rb' ) as f :
                return pickle.load( f )

#class Model ( GalaPyObject ) :
class Model ( ABC ) :
    
    def __init__ ( self ) :
        """ Abstract base interface for all the objects that 
        should be considered models.
        """
        self.params = {}
        
    @abstractmethod
    def set_parameters ( self ) :
        pass



# External includes:
import configparser
import os

# Internal includes
from galapy.configuration.filesystem import _find_or_create_root_dir

_confstr = f"""
#############################
# GalaPy configuration file #
#############################

# Contains the default and user-defined paths
# to the input data-files and template used in the library
# Note that the system will search data-files in all the 
# paths listed below before trying to download the relevant
# files. Research will be made in order of appearence in this section.
[DATAPATHS]

DEFAULT={{0:s}}

# Add eventual new positions in the filesystem here:
# (the name given to the new position is not relevant)
# NEWPATH={os.path.join('absolute','path','to','new','position')}

# Here define boolean values that affect the generic behaviour 
# of the package. 
# It accepts values as true/false, yes/no, 1/0, on/off
[BOOLEANS]

# This variable decides whether to force overwriting of whatever present in a
# path that has been found to already exist during a download request. 
# Setting this to `true` is done at your own risk. 
FORCE_DOWNLOADS=false

# When set to true it will provide additional informations
# while downloading data-files
VERBOSE_DOWNLOADS=true
"""

def _find_or_create_config_file () :
    """ Searches the conf.ini file, it generates one if none exist

    Returns
    -------
    : str
      the absolute path to the configuration file
    
    Raises
    ------
    OSError
      when the configuration file is not present and it is
      not possible to generate any 
    """
    
    rootdir = _find_or_create_root_dir()
    conf_file = os.path.join( rootdir, 'conf.ini' )
    if not os.path.exists( conf_file ) :
        open( conf_file, "w" ).write( _confstr.format(rootdir) )
    elif os.path.isdir( conf_file ) :
        raise OSError( f"Intended galapy configuration file {conf_file}"
                       " is actually a directory" )

    return conf_file

def configure () :
    """ Finds/Generates the conf.ini file, then it reads it
    and converts it to a dictionary, effectively generating the
    rc-parameters.
    
    Returns
    -------
    : dict
      dictionary containing the configuration of the package
    """
    
    conf = configparser.ConfigParser()
    conf.read( _find_or_create_config_file() )

    ###########################################################
    # Use ConfigParser to generate running-command dictionary #
    ###########################################################

    rcdict = {}
    rcdict[ 'datapath' ] = [ v for _, v in conf.items( 'DATAPATHS' ) ]
    for k in conf[ 'BOOLEANS' ].keys() :
        rcdict[ k.lower() ] = conf[ 'BOOLEANS' ].getboolean( k )

    return rcdict



# External imports
import numpy

# Internal imports
import galapy.internal.globs as GP_GBL
from galapy.internal.data import DataFile

class IGM () :
    
    LymanLimit = 912.

    def __init__ ( self ) :
        import os
        
        # Shorter version for convenience
        self.lLL = self.LymanLimit
        
        # Get parameters from file
        intuple = numpy.genfromtxt( DataFile( *GP_GBL.IGM_Inoue14 ).get_file(), 
                                    unpack = True )
        ( self.line, self.lambda_j, 
          self.ALAFj1, self.ALAFj2, self.ALAFj3, 
          self.ADLAj1, self.ADLAj2 ) = intuple
        self.j_lambda = 1. / self.lambda_j
        
    def LS_LAF ( self, lobs, zobs ) :
        
        opt_depth = numpy.zeros_like(lobs)
        
        for jj, (lj, jl) in enumerate(zip(self.lambda_j, 
                                          self.j_lambda)) :
            
            mask = (lj < lobs) & (lobs < lj*(1+zobs))
            l1 = lobs < 2.2*lj
            l2 = (2.2*lj <= lobs) & (lobs < 5.7*lj)
            l3 = 5.7*lj <= lobs
            m1,m2,m3 = mask&(l1,l2,l3)
            
            opt_depth[m1] += self.ALAFj1[jj]*(jl * lobs[m1])**1.2
            opt_depth[m2] += self.ALAFj2[jj]*(jl * lobs[m2])**3.7
            opt_depth[m3] += self.ALAFj3[jj]*(jl * lobs[m3])**5.5
            
        return opt_depth
    
    def LS_DLA ( self, lobs, zobs ) :
        
        opt_depth = numpy.zeros_like(lobs)
        
        for jj, (lj, jl) in enumerate(zip(self.lambda_j, 
                                          self.j_lambda)) :
            
            mask = (lj < lobs) & (lobs < lj*(1+zobs))
            l1 = lobs < 3.0*lj
            l2 = 3.0*lj <= lobs
            m1,m2 = mask&(l1,l2)
            
            opt_depth[m1] += self.ADLAj1[jj]*(jl * lobs[m1])**2
            opt_depth[m2] += self.ADLAj2[jj]*(jl * lobs[m2])**3
            
        return opt_depth

    def LC_LAF ( self, lobs, zobs ) :
    
        # preparing quantities
        opt_depth = numpy.zeros_like(lobs)
        mask = self.lLL < lobs
        lu = lobs/self.lLL

        # different behaviour depending on source redshift
        if zobs < 1.2 :

            # 1
            m1 = mask & (lobs < self.lLL*(1+zobs))
            opt_depth[m1] += 0.325 * (
                lu[m1]**1.2 - (1+zobs)**(-0.9)*lu[m1]**2.1
            )

            # 2 not necessary, return
            return opt_depth

        if (1.2 <= zobs)&(zobs < 4.7) :

            # 1
            m1 = mask & (lobs < 2.2*self.lLL)
            opt_depth[m1] += (
                2.55e-2 * (1+zobs)**1.6 * lu[m1]**2.1 +
                0.325 * lu[m1]**1.2 - 0.250 * lu[m1]**2.1
            )

            # 2
            m2 = mask & ((2.2*self.lLL <= lobs) & 
                         (lobs < self.lLL*(1+zobs)))
            opt_depth[m2] += 2.55e-2 * (
                (1+zobs)**1.6 * lu[m2]**2.1 - lu[m2]**3.7
            )

            # 3 not necessary, return
            return opt_depth

        if 4.7 <= zobs :

            # 1
            m1 = mask & (lobs < 2.2*self.lLL)
            opt_depth[m1] += (
                5.22e-4 * (1+zobs)**3.4 * lu[m1]**2.1 +
                0.325 * lu[m1]**1.2 - 3.14e-2 * lu[m1]**2.1
            )

            # 2
            m2 = mask & ((2.2*self.lLL <= lobs) & 
                         (lobs < 5.7*self.lLL))
            opt_depth[m2] += (
                5.22e-4 * (1+zobs)**3.4 * lu[m2]**2.1 +
                0.218 * lu[m2]**2.1 - 2.55e-2 * lu[m2]**3.7
            )

            # 3
            m3 = mask & ((5.7*self.lLL <= lobs) & 
                         (lobs < self.lLL*(1+zobs)))
            opt_depth[m3] += 5.22e-4 * (
                (1+zobs)**3.4 * lu[m3]**2.1 - lu[m3]**5.5
            )

            # 4 not necessary, return
            return opt_depth

        # this should never be raised
        raise Exception
        
    def LC_DLA ( self, lobs, zobs ) :
    
        # preparing quantities
        opt_depth = numpy.zeros_like(lobs)
        mask = self.lLL < lobs
        lu = lobs/self.lLL

        # different behaviour depending on source redshift
        if zobs < 2.2 :

            # 1
            m1 = mask & (lobs < self.lLL*(1+zobs))
            opt_depth[m1] += (
                0.211 * (1+zobs)**2 - 
                7.66e-2 * (1+zobs)**2.3 * lu[m1]**(-0.3) -
                0.135 * lu[m1]**2
            )

            # 2 not necessary, return
            return opt_depth

        else :

            # 1
            m1 = mask & (lobs < 3.0*self.lLL)
            opt_depth[m1] += (
                0.634 + 4.7e-2 * (1+zobs)**3 -
                1.78e-2 * (1+zobs)**3.3 * lu[m1]**(-0.3) -
                0.135 * lu[m1]**2 - 0.291 * lu[m1]**(-0.3)
            )

            # 2
            m2 = mask & ((3.0*self.lLL <= lobs) & 
                         (lobs < self.lLL*(1+zobs)))
            opt_depth[m2] += (
                4.7e-2 * (1+zobs)**3 - 
                1.78e-2 * (1+zobs)**3.3 * lu[m2]**(-0.3) -
                2.92e-2 * lu[m2]**3
            )

            # 3 not necessary, return
            return opt_depth

        # this should never be raised
        raise Exception

    def optical_depth ( self, lobs, zobs ) :
        return (
            self.LS_LAF( lobs, zobs ) +
            self.LC_LAF( lobs, zobs ) + 
            self.LS_DLA( lobs, zobs ) +
            self.LC_DLA( lobs, zobs )
        )
    
    def transmission ( self, lobs, zobs ) :
        return numpy.exp( -self.optical_depth( lobs, zobs ) )



""" Plotting utilities for the analysis of the results and to generate easily plots 
from GalaPy structures and functions.
"""
######################################################################################
# External imports
import os
import warnings
import fnmatch
import difflib

# As the sane of mind do
import numpy

# For plotting
import matplotlib.pyplot as plt
from matplotlib.style import available as mpl_sty_av
from matplotlib import get_backend
_mpl_backend = get_backend()

# Matplotlib setup of the GalaPy default layout
plt.rcParams['figure.figsize'] = (7,5)
plt.rcParams['font.size'] = 14
plt.rcParams['lines.linewidth'] = 2.
plt.rcParams['text.latex.preamble'] = r"\usepackage{amsmath}"

# this should select seaborn-v0_8-deep
# in all matplotlib installations
# more recent than version 3.6
plt.style.use(difflib.get_close_matches('seaborn-v0_8-deep', mpl_sty_av)[0]) 
clr = plt.rcParams['axes.prop_cycle'].by_key()['color']

######################################################################################
# Internal imports

import galapy as gp
from galapy.internal.utils import filter_strings, shorten_string
from galapy.sampling.Results import Results
from galapy.analysis.funcs import get_parameters_summary_strings, get_parameters_label_strings

######################################################################################
# Basic functions

# a function for formatting the axes' ticks:
def format_axes_ticks(fig, labelsize=14):
    for i, ax in enumerate(fig.axes):
        ax.tick_params(labelsize=labelsize)

######################################################################################
# Default values of the plots settings

_specs_default = {
    'sed_ax_kw' : {
        'xscale' : 'log', 
        'yscale' : 'log',
        'xlim'   : (1.e+0,1.e+10),
        'ylim'   : (1.e-10, 1.e+10),
        'xlabel' : '$\\lambda\\ [\\,\\AA\\,]$',
        'ylabel' : '$S_\\nu\\ [\\mathrm{mJy}]$',
    },
    'sed_legend_kw' : {
        'l1' : { 'loc' : 'upper right',
                 'fontsize' : 12 },
        'l2' : { 'loc' : 'upper left',
                 'ncol' : 2,
                 'fontsize' : 12 },
    },
    'pms_ax_kw' : {
        'xscale' : 'log',
        'yscale' : 'log',
        'ylim'   : (1.e-2,1.e+1),
        'ylabel' : 'Transmission'
    },
    'residuals_ax_kw' : {
        'xlim' : (1., 1.e+10),
        'ylim' : (-1.0, +1.0),
        'xscale' : 'log',
        'yscale' : 'linear',
        'ylabel' : 'residuals',
    },
    'textbox_kw' : {
        'loc' : 'upper right',
        'borderpad' : 0.2, 
        'frameon' : False, 
        'prop' : {
            'fontsize' : 12,
            'bbox' : {
                'boxstyle' : 'round',
                'facecolor' : 'white',
                'alpha' : 0.7
            },
        },
    },
    'getdist_settings' : { 
        'mult_bias_correction_order' : 1,
        'smooth_scale_2D' : 4, 
        'smooth_scale_1D' : 4, 
        'fine_bins': 128,
        'fine_bins_2D' : 128,
        'contours' : [0.68, 0.95, 0.99],
    },
    'titles_kw' : { 
        'digits' : 2, 
        'stat_type' : 'quantiles', 
        'quantile' : None 
    },
    'marker_kw' : { 'ls' : '--', 'lw' : 1.5, 'color' : 'black' },
    'triangle_kw' : { 
        'contour_colors' : ['gray'],
        'line_args' : [{'lw':2, 'color':'gray'}],
    },
}

def show_default_dict ( name = None ) :
    """ Shows the default dictionary for the kwargs dictionary with name 'name'.
    """
    if name is not None :
        return _specs_default[ name ]
    return _specs_default

######################################################################################
# Functions for plotting SEDs

def sed_layout ( redshift, frame, ax = None, **kwargs ) :
    """ Produces/modifies a matplotlib.axes.Axes instance with the grid a labels for
    plotting SED fluxes.
    
    Parameters
    ----------
    redshift : scalar
      the redshift of the observed frame
    frame : str
      one among ('obs', 'rest', 'both'), whether to scale the x-axis
      to the observed frame, the rest-frame or both.
      In the latter case a secondary x-axis will be added on top of the grid
    ax : matplotlib.axes.Axes
      (Optional) an instance of matplotlib axes
    kwargs : dictionary
      (Optional) the eventual keyword arguments to pass to the ```set``` function
      of ```ax```.
    
    Returns
    -------
    : matplotlib.axes.Axes
      the generated axes
    """
    
    # If not passed, attach to the latest axis or generate new one
    if ax is None :
        ax = plt.gca()
    if hasattr(ax, 'sed_layout_set' ) :
        if ax.sed_layout_set :
            return(ax)
        
    # Add secondary axis REST-FRAME WAVELENGHT
    if frame == 'both' :
        def red(x):
            return x * (1+redshift)
        def blue(x):
            return x / (1+redshift)
        secax = ax.secondary_xaxis('top', functions=(blue, red))
        _ = secax.set_xlabel('$\\lambda_\\mathrm{rest}\\ [\\,\\AA\\,]$')
        frame = 'obs'
        
    # Set axis parameters
    default_ax_kw = dict( _specs_default['sed_ax_kw'] )
    default_ax_kw['xlabel'] = f'$\\lambda_\\mathrm{{{frame}}}\\ [\\,\\AA\\,]$'
    default_ax_kw.update(kwargs)
    _ = ax.set(**default_ax_kw)

    # Set background colors
    ax.set_facecolor('white')
    
    # Set an attribute to check the method has already been called
    try :
        _ = setattr(ax, 'sed_layout_set', True)
    except :
        raise Error('Something went wrong.')
    
    return(ax)

def _errorbar_with_uplims( xx, yy, ee, lo, ax = None, **kwargs ) :

    if ax is None :
        ax = plt.gca()
    kwargs = dict(kwargs)
    marker = 'o'
    if 'marker' in kwargs.keys() :
        marker = kwargs.pop('marker')
    
    Pdata = ax.errorbar(
        xx[~lo], yy[~lo], ee[~lo],
        label='data',
        marker=marker,
        markerfacecolor='white',
        markeredgewidth=2.0, 
        markersize=8.0,
        elinewidth=2.0, 
        linestyle='none',
        zorder=3,
        **kwargs
    )
    color = None
    if 'c' in kwargs.keys() :
        color = kwargs.pop( 'c' )
    elif 'color' in kwargs.keys() :
        color = kwargs.pop( 'color' )
    else :
        color = Pdata.lines[0].get_color()
    Puplims = ax.errorbar(
        xx[lo], 2*ee[lo], 1*ee[lo],
        uplims=lo[lo],
        label='uplims',
        capsize = 2.0, 
        capthick=2.0,
        elinewidth=2.0,
        marker='none', 
        linestyle='none',
        zorder=3,
        color = color,
        **kwargs
    )
    return Pdata, Puplims

def sed_obs ( ll, ff, ee, lo, redshift = None, frame = 'rest', ax = None, ax_kwargs = {}, return_uplim_label=False ) :

    ax = sed_layout(redshift, frame, ax, **ax_kwargs)

    Pdata, Puplims = _errorbar_with_uplims( ll, ff, ee, lo, ax=ax )
    
    return Pdata

def sed_components ( ll, components,
                     redshift = None, frame = 'rest', ax = None, ax_kwargs = {} ) :
    ax = sed_layout(redshift, frame, ax, **ax_kwargs)
    Pcomp = []
    for k,L in components.items() :
        Pcomp.append(ax.plot(ll, L, '-', lw = 1.5, label=k)[0])
    return Pcomp

def sed_flux ( ll, flux,
               color = 'black', label = 'total', 
               redshift = None, frame = 'rest', ax = None, ax_kwargs = {} ) :
    ax = sed_layout(redshift, frame, ax, **ax_kwargs)
    return ax.plot( ll, flux, ls='-', color=color, lw=1.5, label=label)[0]

def sed_1sigma2sigma ( ll, flux, err, center = False, color = 'gray', center_label = 'mean', 
                       redshift = None, frame = 'rest', ax = None, ax_kwargs = {} ) :
    ax = sed_layout( redshift, frame, ax, **ax_kwargs )
    handles = []
    if center :
        handles.append( ax.plot( ll, flux,
                                 ls='-', color=color,
                                 lw=1.5, label=center_label )[0] )
    handles.append(
        ax.fill_between(ll, flux-err, flux+err,
                        label='$1$-$\\sigma$',
                        edgecolor='none', color=color, alpha=0.5)
    )
    handles.append(
        ax.fill_between(ll, flux-2*err, flux+2*err,
                        label='$2$-$\\sigma$',
                        edgecolor='none', color=color, alpha=0.2)
    )
    return handles

def sed_flux_res ( res,
                   model = None,
                   observation = None,
                   plot_observation = False,
                   plot_components = False, 
                   plot_contours = True,
                   frame = 'both',
                   show_legend = True,
                   legend_kwargs = {},
                   ax = None, 
                   ax_kwargs = {}) :
    """ Plots the formatted SED flux from a galapy.sampling.Results instance.
    Optional specs that can be activated or de-activated are the observed points,
    1- and 2-sigma contours around the mean of the sampled SEDs and the different
    physical components contributing to the total best-fit SED.
    
    Parameters
    ----------
    res : Results instance
        A ``Results`` instance from a sampling run.
    model : GXY instance
        (Optional) A galaxy model or any class derived from ``galapy.Galaxy.GXY``.
    observation : Observation instance
        (Optional) an ``Observation`` instance. Note that if ``plot_observation=False``
        it will be ignored anyways.
    plot_observation : bool
        Whether to plot the observational data used in the sampling (dafault is ``False``).
    plot_components : bool
        Whether to plot the different components contributing to the best-fit SED 
        (dafault is ``False``).
    plot_contours : bool
        Whether to plot the 1- and 2-sigma contours around the mean SED (dafault is ``False``).
    frame : string
        One among {``'rest'``, ``'obs'``, ``'both'``}, choose the frame, observed- or rest-frame,
        to plot the wavelength-axis (default is ``'both'``)
    show_legend : bool
        (Optional)
    legend_kwargs : dict
        (Optional)
    ax : matplotlib.axes.Axes
        (Optional) an instance of matplotlib axes
    ax_kwargs : dict
        (Optional) keyword arguments to pass to the function ``sed_layout``
    
    Returns
    -------
    : matplotlib.axes.Axes
        An instance of ``matplotlib.axes.Axes`` with the plotted fluxes
    """
    
    if not isinstance( res, Results ) :
        raise AttributeError( 
            'Attribute "res" should be an instance of type ``Results``'
        )
        
    if model is None :
        model = res.get_model()
    else :
        if not isinstance( model, gp.Galaxy.GXY ) :
            raise AttributeError(
                'Attribute "model" should be an instance of type ``GXY``'
            )
    if plot_observation :
        if observation is None :
            observation = res.get_observation()
        else :
            if not isinstance( observation,
                               gp.sampling.Observation.Observation ) :
                raise AttributeError(
                    'Attribute "observation" should be an instance of type ``Observation``'
                )

    # Set the wavelength (lambda) axis:
    # If 'redshift' is a free parameter the local variable
    # will be set to the best-fit value, otherwise the
    # spectroscopic redshift will be extracted from the model.
    redshift = None
    lx = None
    obs_scale = frame in set(['obs', 'both'])
    if 'galaxy.redshift' in res.get_sampling_params().par_free :
        redshift = res.get_bestfit( 'params' )['redshift']
        lx = model.wl()
        if obs_scale :
            lx *= ( 1 + redshift )
    else :
        redshift = model.redshift
        lx = model.wl( obs = obs_scale )
        
    # Set the layout
    ax = sed_layout( redshift, frame, ax = ax, **ax_kwargs )
    
    ##################################################################
    # Here the plots start
    
    legend_primary   = []
    legend_secondary = []
    if show_legend :
        legend_kw = dict( _specs_default['sed_legend_kw'] )
        legend_kw.update( legend_kwargs )
        
    
    # Plot the observational dataset
    if plot_observation :
        ll, ff, ee, lo = ( observation.pms.lpiv,
                           observation.fluxes,
                           observation.errors,
                           observation.uplims )
        legend_primary.append(
            sed_obs( ll, ff, ee, lo,
                     redshift = redshift,
                     frame = frame,
                     ax=ax )
        )
        
    # Plot the different components
    flux_bestf = None
    if plot_components :
        bestfit_p = res.get_bestfit( 'params' )
        model.set_parameters(**bestfit_p)
        flux_bestf = model.get_SED()
        components = model.components_to_flux()
        legend_secondary += sed_components( lx, components,
                                            redshift = redshift,
                                            frame = frame,
                                            ax=ax )
        
    if flux_bestf is None :
        flux_bestf = res.get_bestfit( 'SED' )
    legend_primary.append( sed_flux( lx, flux_bestf, label = 'best-fit',
                                     redshift = redshift,
                                     frame = frame,
                                     ax=ax ) )
        
    if plot_contours :
        flux_mean = res.get_mean( 'SED' )
        flux_err  = res.get_std( 'SED' )
        legend_primary += sed_1sigma2sigma( lx, flux_mean, flux_err,
                                            redshift = redshift,
                                            frame = frame,
                                            ax=ax )
    
    ##################################################################
    # Finalise and return
    
    # LEGEND:
    if show_legend :
        lpri = ax.legend(handles = legend_primary, **legend_kw['l1'] )
        ax.add_artist(lpri)
        if len(legend_secondary)>0 :
            lsec = ax.legend(handles = legend_secondary, **legend_kw['l2'] )
            ax.add_artist(lsec)
        
    return(ax)


def sed_residuals_res ( res,
                        frame = 'both', 
                        plot_contours = False, 
                        plot_chi2 = True,
                        ax = None,
                        text_kwargs = {},
                        ax_kwargs = {}) :
    """
    Plots the formatted residuals with respect to the best-fit model from a sampling run.
    
    Parameters
    ----------
    res : Results instance
        A ``Results`` instance from a sampling run.
    frame : string
        One among {``'rest'``, ``'obs'``, ``'both'``}, choose the frame, observed- or rest-frame,
        to plot the wavelength-axis (default is ``'both'``)
    plot_contours : bool
        Whether to plot the 1- and 2-sigma contours around the mean SED (dafault is ``False``).
    plot_chi2 : bool
        Whether to plot a text box with the best-fit value of the reduced 
        chi2 (dafault is ``True``).
    ax : matplotlib.axes.Axes
        (Optional) an instance of matplotlib axes
    text_kwargs : dict
        (Optional) keyword arguments to pass to the matplotlib.offsetbox.AnchoredText class
        (regulates shape and text-formatting of the text-box for the eventual chi2 plot)
    ax_kwargs : dict
        (Optional) keyword arguments to pass to the function ``sed_layout``
    
    Returns
    -------
    : matplotlib.axes.Axes
        An instance of ``matplotlib.axes.Axes`` with the plotted fluxes
    """
    
    # Get sampled model
    pbest = res.get_bestfit('params')
    pgxy = res.get_model()
    pgxy.set_parameters(**pbest)
    redshift = pgxy.redshift

    # Get reference observation
    obs = res.get_observation()
    ll, ff, ee, lo = ( obs.pms.lpiv,
                       obs.fluxes,
                       obs.errors,
                       obs.uplims )
    lsed = pgxy.wl(obs=(frame != 'rest'))

    # compute residuals
    chi = res.get_residuals()
    
    # Set axes
    ax_kw = dict(_specs_default['residuals_ax_kw'])
    ax_kw.update(ax_kwargs)
    ax = sed_layout(redshift, frame=frame, ax = ax, **ax_kw )
    
    #
    if plot_contours :
        sedstd = res.get_std('SED')
        wnstd0 = sedstd > 0.0
        _ = sed_1sigma2sigma( lsed[wnstd0], 
                              (res.get_mean('SED')-res.get_bestfit('SED'))[wnstd0]/sedstd[wnstd0], 
                              numpy.ones_like(lsed[wnstd0]),
                              center=True, redshift=redshift, frame=frame, ax = ax)
        
    _ = sed_flux( lsed, numpy.zeros_like(lsed), redshift=redshift, frame=frame, ax = ax )
    Pdata, *_ = _errorbar_with_uplims(
        ll[~lo], chi[~lo],
        numpy.zeros_like(ee[~lo]),
        numpy.zeros_like(lo[~lo]),
        ax = ax,
    )
    obs_color = Pdata.lines[0].get_color()
    _ = _errorbar_with_uplims(
        ll[lo],
        numpy.zeros_like(chi[lo]),
        numpy.zeros_like(ee[lo]),
        numpy.zeros_like(lo[lo]),
        ax = ax,
        color = obs_color,
        marker = 'x'
    )
    
    # Plot eventual text-box
    if plot_chi2 :
        from matplotlib.offsetbox import AnchoredText
        text_kw = dict(_specs_default['textbox_kw'])
        text_kw.update(text_kwargs)
        
        textstr = f'$\\chi^2_\\mathrm{{red}} = {numpy.sum(chi**2)/(len(ll)-res.ndim):.2f}$'
        text = AnchoredText(textstr, **text_kw)
        _ = ax.add_artist(text)
        
    return ax

######################################################################################
# Functions for plotting the photometric system

def photometric_system ( obj, colors = None, ax = None, ax_kwargs = {} ) :
    """ Generates a visualisation of the photometric system bandpass transmissions.
    
    Parameters
    ----------
    obj : object
        An instance of ``galapy.PhotometricSystem.PMS`` or any other object with a 
        ``pms`` attribute which is an instance of ``galapy.PhotometricSystem.PMS``
        (e.g. ``galapy.Galaxy.GXY`` or ``galapy.Observation.OBS``)
    colors : iterable
        (Optional) a list or iterable of valid colors. One color per bandpass is necessary
    ax : matplotlib.axes.Axes
        (Optional) an instance of matplotlib axes
    ax_kwargs : dictionary
        (Optional) the eventual keyword arguments to pass to the ``set`` function
        of ``ax``.

    Returns
    -------
    : matplotlib.axes.Axes
    an instance of matplotlib axes
    """
    from galapy.PhotometricSystem import PMS
    
    # Check input object (it should be or contain a PMS instance)
    if not isinstance(obj, PMS) :
        if hasattr(obj, 'pms') and isinstance( obj.pms, PMS ) :
            pms = obj.pms
        else :
            raise AttributeError( 'Input object ``obj`` must be an instance of '
                                  'either ``galapy.PhotometricSystem.PMS``, '
                                  'or any other object with a ``pms`` attribute '
                                  'which an instance of ``galapy.PhotometricSystem.PMS`` '
                                  '(e.g. ``galapy.Galaxy.GXY`` or ``galapy.Observation.OBS``)' )
    else :
        pms = obj
    
    # generate or check color iterable
    if colors is None :
        colors = plt.cm.plasma(numpy.linspace(0.,1.,len(pms.keys)))
    elif len(colors) != len(pms) :
        raise AttributeError( 'Attribute ``color`` must be a list or iterable of valid colors '
                              'with the same length of the photometric system' )
            
    # If not passed, attach to the latest axis or generate new one
    if ax is None :
        ax = plt.gca()
        
    # Set axes properties
    default_ax_kw = dict( _specs_default['pms_ax_kw'] )
    default_ax_kw.update(ax_kwargs)
    _ = ax.set(**default_ax_kw)
    xax2 = ax.secondary_xaxis('top', functions=(lambda x : x, lambda x : x))
    _ = xax2.set( xscale = default_ax_kw['xscale'], xlabel = 'wavelength $[\\AA]$' )

    # Set background color
    ax.set_facecolor('white')
    
    # Plot
    xticks = []
    xticklabels = []
    for k, c in zip( pms.keys, colors ) :
        v = pms.bpt[k]
        xticks += [v.get_lpiv()]
        xticklabels += [shorten_string(k)]
        ax.plot( v.get_xaxis(), numpy.asarray( v.get_xaxis() )*numpy.asarray( v.get_yaxis() ), color = c)
        ax.axvline( v.get_lpiv(), color = c, ls = '--', lw = 2.)
    _ = ax.set_xticks(xticks)
    _ = ax.set_xticklabels(xticklabels, rotation=90, fontsize = 11)

    return ax

######################################################################################
# Functions for plotting posteriors

def corner_res ( res, handler = None, which_params = None, getdist_settings = None,
                 param_limits = 'auto', plot_titles = True, mark = 'bestfit',
                 titles_kw = {}, triangle_kw = {}, marker_kw = {} ) :
    """ Plot the triangle plot with the 2D posteriors of a sampling run 
    and 1D marginals on the diagonal.
    
    Parameters
    ----------
    res :  Results instance
        A ``Results`` instance from a sampling run.
    handler : ModelParameters instance
        (Optional) The ``ModelParameters`` corresponding to the given sampling run
    which_params : str or sequence of str
        Either a single string or a sequence of strings. Name of the parameters to show on the 
        triangle plot. Also accepts wildcards (e.g. ``which_params = 'sfh*'`` will show all the
        parameters that contain the sub-string ``sfh``).
    getdist_settings : dict
        a dictionary of custom analysis settings to pass to the ``getdist.MCSamples`` 
        settings argument 
        (for further informations see the documentation of ``getdist``)
        (to see defaults call ``galapy.analysis.plots.show_default_dict('getdist_settings')``)
    param_limits : str or sequence or dict
        If a string is passed it has to be one among

        - 'auto' : set the axes limits automatically
        - 'prior' : set the axes to the limits of the prior

        Otherwise it can be a 

        - 2D sequence (list or tuple) with dimensions (ndim, 2), where
          ndim is the number of free parameters of the sampling run. Each of the
          ndim couples will be assigned to the ordered list of free-parameters.
        - a dictionary where the keys are the names of the free parameters of the
          sampling run and the values are sequences of length = 2 with the limits
          for the corresponding free parameter.

    plot_titles : bool
        Whether to plot titles above the diagonal marginal posteriors with summary
        statistics computed from the corresponding posteriors (default summary is 
        the median with 68% uncertainty, to change this behaviour modify the 
        argument ``titles_kw``)
    mark : str
        What position in the posteriors space to highlight with lines, 
        default is the 'bestfit', other available positions are 
        'mean' and 'median'.
    titles_kw : dict
        dictionary to modify the titles above the diagonal marginal posteriors.
        It will be passed to function ``galapy.analysis.funcs.get_parameters_summary_strings()``
        (to see defaults call ``galapy.analysis.plots.show_default_dict('titles_kw')``)
    triangle_kw : dict
        Dictionary to modify the aspect of the triangle plot, these are the
        keyword arguments passed to ``triangle_plot()`` function of ``getdist``
        (to see defaults call ``galapy.analysis.plots.show_default_dict('triangle_kw')``)
    marker_kw : dict
        Dictionary to modify the aspect of the markers (color, linewidth, ...)
        (to see defaults call ``galapy.analysis.plots.show_default_dict('marker_kw')``)
    
    Returns
    -------
    : matplotlib.pyplot.Figure instance
    : matplotlib.axes.Axes instance
    """
    from getdist import plots, MCSamples
    import getdist
    # necessary to fix the hard-coded matplotlib.use('Agg') forced by getdist
    plt.switch_backend(_mpl_backend)

    ############################################################################
    # Set default kwarg-dictionaries
    
    default_getdist_settings = dict( _specs_default['getdist_settings'] )
    default_titles_kw        = dict( _specs_default['titles_kw'] )
    default_marker_kw        = dict( _specs_default['marker_kw'] )
    default_triangle_kw      = dict( _specs_default['triangle_kw'] )
    
    ############################################################################
    # Check inputs
    
    if not isinstance( res, Results ) :
        raise AttributeError( 
            'Attribute "res" should be an instance of type ``Results``'
        )
    
    if handler is None :
        handler = res.get_sampling_params()
    else :
        if not isinstance( handler, gp.Handlers.ModelParameters ) :
            raise AttributeError(
                'Attribute "handler" should be an instance of type ``ModelParameters``'
            )
    
    if which_params is None :
        which_params = handler.par_free
    else :
        try :
            which_params = filter_strings( handler.par_free, which_params )
        except ValueError :
            raise TypeError(
                "Argument ``which_params`` must be a string, a list of strings or tuple of strings"
           )
        
    if getdist_settings is None :
        getdist_settings = {}
    default_getdist_settings.update(getdist_settings)
    
    if isinstance(param_limits, str) :
        if param_limits not in ('auto', 'prior') :
            raise ValueError(
                "Allowed string-values for argument ``param_limits`` are 'auto' and 'prior'"
            )
    elif isinstance(param_limits, MM) :
        pass
    elif isinstance(param_limits, (list, tuple)) :
        param_limits = dict(zip(handler.par_free, param_limits))
    else :
        raise AttributeError(
            "Argument ``param_limits`` must be a string or a dictionary"
        )

    markers = {}
    if mark is not None :
        if mark == 'bestfit' :
            markers = dict(zip(handler.par_free, res.get_bestfit('samples')))
        elif mark == 'mean' :
            markers = dict(zip(handler.par_free, res.get_mean('samples')))
        elif mark == 'median' :
            markers = dict(zip(handler.par_free, res.get_quantile('samples')))
        else :
            warnings.warn( "Argument ``mark`` should be one among 'bestfit', 'median', 'mean'." )
    
    # Update keyword arguments dictionaries
    default_triangle_kw.update(triangle_kw)
    default_titles_kw.update(titles_kw)
    default_marker_kw.update(marker_kw)
    
    ############################################################################
    # Prepare samples for plotting
    
    # Prepare dictionary with Prior-limits
    ranges = dict(zip(handler.par_free, handler.par_prior))
    
    # Prepare list with labels
    labels = get_parameters_label_strings( handler )
    
    # Extract the sampler's name from results
    sampler = None
    if res.sampler == 'dynesty' :
        sampler = 'nested'
    if res.sampler == 'emcee' :
        sampler = 'mcmc'
    
    # Build Samples-object
    mcsamples = MCSamples(
        samples = res.samples,
        loglikes = -res.logl,
        weights = res.weights,
        ranges = ranges,
        settings = default_getdist_settings,
        names = handler.par_free, 
        sampler = sampler,
        labels = list(labels.values())
    )
    
    # Prepare dictionary with parameters limits
    if param_limits == 'auto' : param_limits = {}
    if param_limits == 'prior' : param_limits = ranges 
     
    ############################################################################
    # Start plotting
    
    # Make triangle plot
    g = plots.get_subplot_plotter()
    g.settings.axis_tick_max_labels = 3
    g.settings.num_plot_contours = len(default_getdist_settings)
    g.triangle_plot(
        [mcsamples], params=which_params, 
        filled=True, title_limit=0, 
        param_limits = param_limits,
        markers = markers,
        marker_args = default_marker_kw,
        **default_triangle_kw
    )

    # Extract matplotlib objects
    fig, axes = g.fig, g.subplots
    
    # Set marker-args also on marginals:
    if len(markers) > 0 :
        for i in range(len(axes)) :
            axes[i,i].get_lines()[1].set(**default_marker_kw)
    
    # Plot formatted titles if requested
    if plot_titles :
        title_dict = dict(zip(
            handler.par_free,
            get_parameters_summary_strings( res, **default_titles_kw )
        ))
        for i, (ax, key) in enumerate(zip(axes, which_params)) :
            ax[i].set_title( ' '.join(('$', labels[key], '=', title_dict[key], '$')),
                             fontsize = g.settings.axes_fontsize )

    # Axes-ticks formatting
    format_axes_ticks( fig, labelsize = g.settings.axes_fontsize )

    ############################################################################
    # done and dusted.
    
    return fig, axes

######################################################################################



# External imports
import os
import warnings
import requests
import urllib.parse
import tarfile
import shutil
import argparse

# Internal imports
from galapy.configuration import rcParams
from galapy.configuration.filesystem import _find_or_create_root_dir
import galapy.internal.globs as GP_GBL

################################################################################

def is_url ( string ) :
    """
    Test whether a string is a valid URL for :func:`download_file`.

    Parameters
    ----------
    string : str
        The string to test.

    Returns
    -------
    status : bool
        String is URL or not.

    Notes
    -----
    This routine has been adapted from the Astropy Collaboration (http://www.astropy.org)
    """
    u = urllib.parse.urlparse(string)
    # we can't just check that url.scheme is not an empty string, because
    # file paths in windows would return a non-empty scheme (e.g. e:\\
    # returns 'e').
    return u.scheme.lower() in ['http', 'https', 'ftp', 'sftp', 'ssh', 'file']

################################################################################

def download_file ( url, out_file, overwrite = False, verbose = True, get_kw = {} ) :
    """ Download file from known url into defined position

    Parameters
    ----------
    url : str
      The complete url of the file to download
    out_file : str
      Absolute path to output file
    overwrite : bool
      If set to True, the eventually already existing out_file is overwritten
    verbose : bool
      If set to True, prints additional informations while processing

    Returns
    -------
    : int
    size in bytes written to the out_file

    Raises
    ------
    OSError 
      If the input url string is not a valid URL. 
      If the output file directory does not exist and the system is not able to create it.
    requests.exceptions.HTTPError
      If an HTTP error occurred.
    Exception
      If some other error occurs.
    """

    if not is_url( url ) :
        raise OSError( f"The string provided {url:s} is not a valid URL." )

    # Extracting bytes from the provided URL:
    _bytes = b''
    try :
        response = requests.get(url, **get_kw)
        response.raise_for_status()
    except requests.exceptions.HTTPError as http_err :
        print( f'HTTP error occurred: {http_err}' )
    except Exception as err :
        print( f'Other error occurred: {err}' )
        raise
    else :
        _bytes = response.content

    # If overwrite is set to False
    # and the path already exist leave the file as is
    # returns zero (number of bytes written)
    if os.path.exists( out_file ) and not overwrite :
        warnings.warn( f"File {out_file} already exists." )
        return 0
        
    # Checking whether the directory for file output is accessible 
    out_dir = os.path.dirname( out_file )
    if not os.path.isdir( out_dir ) :
        if verbose : print( f"Directory {out_dir} does not exist, it will be created." )
        try :
            # creates multi-level subdirs. similarly to the *Nix command `mkdir -p`
            # (while os.mkdir() only allows to create the highest level directory)
            os.makedirs( out_dir ) 
        except OSError:
            print ( f"Creation of the directory {out_dir} failed" )
            raise
        else :
            if verbose : print( f"Successfully created the directory {out_dir}" )

    return open(out_file, "wb").write(_bytes)

################################################################################

class DataFile () :
    """
    - search in the rcParams[ 'datapath' ] list
    - if not present, download and store it 
    """

    def __init__ ( self, file_name, path_line ) :

        # allocate variable to store the file-path
        self._filepath = None

        # search in the available data-paths for the requested file
        for datapath in rcParams[ 'datapath' ] :
            _testpath = os.path.join( datapath, *path_line, file_name )
            if not os.path.exists( _testpath ) :
                continue
            elif os.path.isdir( _testpath ) :
                raise OSError( f'The DataFile {file_name} '
                               f'in position {os.path.join(datapath, *path_line)} '
                               'is a directory.' )
            else :
                self._filepath = _testpath

        # if the data-file is not present in the system download it from remote
        if self._filepath is None :
            where = os.path.join( rcParams[ 'datapath' ][ 0 ],
                                  *path_line,
                                  file_name )
            try :
                byte_count = download_file(
                    url = GP_GBL.DATA_URL.format( '/'.join( [ *path_line, file_name ] ) ),
                    out_file = where,
                    overwrite = rcParams[ 'force_downloads' ],
                    verbose = rcParams[ 'verbose_downloads' ]
                )
            except Exception as err :
                print( f'An error occurred: {err}' )
                raise
            else :
                if rcParams[ 'verbose_downloads' ] : print( f'Downloaded {byte_count} bytes.' )
                self._filepath = where
            

    def get_file ( self ) :
        """ 
        - return the string with the cross-platform path
        """
        return self._filepath

################################################################################
# def search_file () :
#     pass
################################################################################

def download_database ( loc = None, name = None, version = None,
                        database = None, overwrite = False, verbose = True ) :
    """ A function for downloading the database from a github repository.
    (Or from any URL that can be converted into JSON and contains a 'tarball_url'
     attribute with the complete url to the tarball to be downloaded)

    Parameters:
    -----------
    loc : str
        Location in the filesystem where to download and extract the database.
        Default is None and will download in the default position specified in
        galapy.rcParams
    name : str
        Name to assign to the directory inside loc where the database will be extracted
        Default is None and will be set to the default name in galapy.internal.globs.DATA_DIR
    version : str
        The database version to download (Default is None and will download the latest version)
    database : str
        A valid URL to a database. Typically this would be in the form
        ``'https://api.github.com/repos/<user>/<repository>/releases'``.
        Any URL that can be converted by function requests.request.json() to a dictionary
        containing a valid URL to a tarball associated to key 'tarball_url' is valid.
    overwrite : bool
        Whether to overwrite whatever present at position ``os.path.join(loc, name)``.
        Default is ``False``, changing this SHOULD BE DONE AT OWN RISK.
    verbose : bool
        Whether to download verbosely. Default is ``True``

    Returns:
    --------
    : None
    """
    # a future implementation should give the possibility to add a new location to
    # the rcParams permanently
    from json import JSONDecodeError

    if loc is None :
        loc = rcParams['datapath'][-1]
    else :
        # checks the path already exists:
        if not os.path.exists( loc ) :
            # first, if the path is not present try to make a new directory:
            try :
                os.mkdir( loc )
            except OSError :
                if not os.path.isdir( loc ) :
                    raise
        elif not os.path.isdir( loc ) :
            # if the path exists but it is not a directory raise
            raise OSError( f'Intended galapy root directory {loc} is actually a file' )

        # return absolute path to selected location
        loc = os.path.abspath( loc )
        warnings.warn( f'The database will be downloaded in {loc}. '
                       'To add the custom location permanently to your '
                       'search paths modify the file '
                       f'{os.path.join(_find_or_create_root_dir(), "conf.ini")}' )
        
    if version is None :
        version = 'latest'
    elif version == 'latest' :
        pass
    elif version == 'default' :
        version = f'v{DATA_VERSION}'
    else :
        version = f'v{version}'

    if name is None :
        name = GP_GBL.DATA_DIR        
    destpath = os.path.join( loc, name )
    # If overwrite is set to False
    # and the path already exists leave the file as is
    # and return
    if os.path.exists( destpath ) and os.path.isdir( destpath ) :
        if not overwrite :
            warnings.warn( f"Directory {destpath} already exists." )
            return;
        else :
            shutil.rmtree( destpath )
    elif os.path.exists( destpath ) and not os.path.isdir( destpath ) :
        if not overwrite :
            warnings.warn( f"{destpath} is an already existing File." )
            return;
        else :
            os.remove( destpath )    

    # Check on input variable database
    if database is None :
        database = GP_GBL.DATABASE
    elif not is_url( database ) :
        raise OSError( f"The database provided {database:s} is not a valid URL." )

    #######################################################################
    # Download database

    # Request for a given release and check the URL is correct and working
    release = requests.get( database )
    try :
        release.raise_for_status()
    except requests.exceptions.HTTPError as http_err :
        print( f'HTTP error occurred: {http_err}' )
        raise
    except Exception as err :
        print( f'Other error occurred: {err}' )
        raise

    # converting json from list to dict
    releaselist = release.json()
    releasedict = {'latest' : releaselist[0]}
    releasedict.update(
        { rel['tag_name'] : rel for rel in releaselist }
    )

    # Extract from JSON the tarball URL and check for exceptions
    try :
        tarurl = releasedict[version]['tarball_url']
    except JSONDecodeError as err :
        print( f'Provided URL cannot be converted to JSON dictionary' )
        raise
    except Exception as err :
        print( f'An error occurred: {err}' )
        raise

    # Try downloading the tarball and eventually raise exception.
    where = os.path.join( loc, f'{name}.tar.gz' )
    try :
        if rcParams[ 'verbose_downloads' ] :
            print( 'downloading tarball, this might require some time ...' )
        nbytes = download_file( tarurl, where,
                                overwrite = overwrite,
                                verbose = verbose )
    except Exception as err :
        print( f'An error occurred: {err}' )
        raise
        
    if rcParams[ 'verbose_downloads' ] : print( f'Downloaded {nbytes} bytes' )

    #######################################################################
    # Extract and rename the archive
    
    # Open tarball
    archive = tarfile.open(where)

    # Get common prefix name (e.g. galapy_database-1.0.0)
    archive_root = os.path.commonprefix( archive.getnames() )

    # Extract everything from tarball on assigned position
    archive.extractall( loc )

    # Close tarball
    archive.close()

    # Rename extracted folder to assigned name
    os.rename(
        os.path.join( loc, archive_root ), # previous name
        destpath                           # new name
    )
    
    return;

################################################################################

def _entrypoint_download_database () :

    ####################################################################
    # Read command-line arguments:
    
    parser = argparse.ArgumentParser( description = 'options' )
    parser.add_argument( '--location', '-l',
                         dest = 'loc',
                         type = str,
                         default = None,
                         help = (
                             'Location in the filesystem where to download and ' +
                             'extract the database. ' +
                             f'DEFAULT: {rcParams["datapath"][-1]}'
                         ) )
    parser.add_argument( '--name', '-n',
                         dest = 'name',
                         type = str,
                         default = 'galapy_database',
                         help = (
                             'Name to assign to the directory inside ``loc`` where ' +
                             'the database will be extracted. ' +
                             'DEFAULT: "galapy_database"'
                         ) )
    parser.add_argument( '--version', '-v',
                         dest = 'version',
                         type = str,
                         default = 'latest',
                         help = (
                             'The database version to download. ' +
                             'DEFAULT: "latest"'
                         ) )
    args = parser.parse_args()
    
    ####################################################################

    return download_database(
        loc = args.loc,
        name = args.name,
        version = args.version,
        database = GP_GBL.DATABASE,
        overwrite = False,
        verbose = True
    )

################################################################################



import setuptools
import os
import sysconfig
import numpy as np

from setuptools import setup, find_packages

################################################################################
# Functions for reading the version 

def read( rel_path ):
    here = os.path.abspath( os.path.dirname( __file__ ) )
    with open( os.path.join( here, rel_path ), 'r' ) as fp :
        return fp.read()

def get_version( rel_path ):
    for line in read( rel_path ).splitlines() :
        if line.startswith( '__version__' ) :
            delim = '"' if '"' in line else "'"
            return line.split( delim )[ 1 ]
    else:
        raise RuntimeError( "Unable to find version string." )


################################################################################
# Functions for building the library

def return_extensions () :
    from pybind11.setup_helpers import Pybind11Extension
    
    #############################################################################
    # C++ implementation of the interpolation class

    ext_intp = Pybind11Extension(
        "galapy.internal.interp",
        sorted(
            [ os.path.join( 'pybind11', 'pyb11_interpolation.cpp' ) ]
        ),
        include_dirs = sorted( [ os.path.join( 'c++', 'utl', 'include' ),
                                 os.path.join( 'pybind11' ) ] ),
        libraries = [ "m" ],
        extra_compile_args=[ '-std=c++17' ]
    )
    
    #############################################################################
    # C++ implementation of SFH functions and types

    ext_sfh = Pybind11Extension(
        "galapy.SFH_core",
        sorted(
            [ os.path.join( 'pybind11', 'pyb11_CSFH.cpp' ),
              os.path.join( 'c++', 'sfh', 'src', 'sfh_base.cpp' ),
              os.path.join( 'c++', 'sfh', 'src', 'sfh_insitu.cpp') ]
        ),
        include_dirs = sorted( [ os.path.join( 'c++', 'sfh', 'include'),
                                 os.path.join( 'c++', 'imf', 'include'),
                                 os.path.join( 'c++', 'utl', 'include' ),
                                 os.path.join( 'pybind11' ) ] ),
        libraries = [ "m" ],
        extra_compile_args=[ '-std=c++17' ]
    )

    #############################################################################
    # C++ implementation of CSP functions and types

    ext_csp = Pybind11Extension(
        "galapy.CSP_core",
        sorted(
            [ os.path.join( 'pybind11', 'pyb11_CCSP.cpp' ),
              os.path.join( 'c++', 'csp', 'src', 'csp.cpp' ) ]
        ),
        include_dirs = sorted( [ os.path.join( 'c++', 'csp', 'include' ),
                                 os.path.join( 'c++', 'utl', 'include' ),
                                 os.path.join( 'pybind11' ) ] ),
        libraries = [ "m" ],
        extra_compile_args=[ '-std=c++17' ]
    )

    #############################################################################
    # C++ implementation of ISM functions and types

    ext_ism = Pybind11Extension(
        "galapy.ISM_core",
        sorted(
            [ os.path.join( 'pybind11', 'pyb11_CISM.cpp' ),
              os.path.join( 'c++', 'ism', 'src', 'ism.cpp' ) ]
        ),
        include_dirs = sorted( [ os.path.join( 'c++', 'ism', 'include' ),
                                 os.path.join( 'c++', 'utl', 'include' ),
                                 os.path.join( 'pybind11' ) ] ),
        libraries = [ "m" ],
        extra_compile_args=[ '-std=c++17' ]
    )

    #############################################################################
    # C++ implementation of the Nebular Free-Free emission (Bremsstrahlung)

    ext_nff = Pybind11Extension(
        "galapy.NFF_core",
        sorted(
            [ os.path.join( 'pybind11', 'pyb11_CNFF.cpp' ) ]
        ),
        include_dirs = sorted( [ os.path.join( 'c++', 'rad', 'include' ),
                                 os.path.join( 'c++', 'utl', 'include' ),
                                 os.path.join( 'pybind11' ) ] ),
        libraries = [ "m" ],
        extra_compile_args=[ '-std=c++17' ]
    )

    #############################################################################
    # C++ implementation of the Synchrotron emission (generic parametric)

    ext_syn = Pybind11Extension(
        "galapy.SYN_core",
        sorted(
            [ os.path.join( 'pybind11', 'pyb11_CSYN.cpp' ) ]
        ),
        include_dirs = sorted( [ os.path.join( 'c++', 'rad', 'include' ),
                                 os.path.join( 'c++', 'utl', 'include' ),
                                 os.path.join( 'pybind11' ) ] ),
        libraries = [ "m" ],
        extra_compile_args=[ '-std=c++17' ]
    )    

    #############################################################################
    # C++ implementation of BPT functions and types

    ext_bpt = Pybind11Extension(
        "galapy.BandpassTransmission",
        sorted(
            [ os.path.join( 'pybind11', 'pyb11_transmission.cpp' ) ]
        ),
        include_dirs = sorted( [ os.path.join( 'c++', 'utl', 'include' ),
                                 os.path.join( 'pybind11' ) ] ),
        libraries = [ "m" ],
        extra_compile_args=[ '-std=c++17' ]
    )

    return [
        ext_intp,
        ext_sfh,
        ext_csp,
        ext_ism,
        ext_nff,
        ext_syn,
        ext_bpt
    ]

def main():
    
    #############################################################################
    # Call setup
    
    setup( name        = "galapy-fit",
           version     = get_version( os.path.join('galapy', '__init__.py') ),
           description = "GalaPy - Spectral modelling tool for galaxies in Python",
           package_dir = {
               'galapy' : 'galapy',
               'galapy.sampling' : os.path.join( 'galapy', 'sampling' ),
               'galapy.configuration' : os.path.join( 'galapy', 'configuration' ),
               'galapy.internal' : os.path.join( 'galapy', 'internal' ),
               'galapy.analysis' : os.path.join( 'galapy', 'analysis' ),
               'galapy.io' : os.path.join( 'galapy', 'io' ),
           },
           packages = [ 'galapy',
                        'galapy.configuration',
                        'galapy.internal',
                        'galapy.sampling',
                        'galapy.analysis',
                        'galapy.io' ],
           ext_modules = return_extensions(),
           include_package_data = True,
           entry_points = {
               'console_scripts' : [
                   'galapy-fit = galapy.sampling.Run:_run',
                   'galapy-genparams = galapy.sampling.Run:_generate_parameter_file',
                   'galapy-download-database = galapy.internal.data:_entrypoint_download_database',
               ]
           },
           install_requires = [
               'pybind11',
               'numpy',
               'scipy',
               'tqdm',
               'emcee',
               'dynesty',
               'matplotlib',
               'getdist',
               'requests',
               'h5py',
               'pytest'
           ]
    )

if __name__ == "__main__":
    main()



""" GalaPy module for combining Simple Stellar Populations (SSP) into a 
Composite Stellar Population (CSP).
"""

# External imports
import numpy
import os
import warnings

# Internal imports
from .CSP_core import loadSSP as _loadSSP, CCSP
from .SYN_core import CSYN
import galapy.internal.globs as GP_GBL
from galapy.internal.data import DataFile

def _recursive_list_ssp_libs ( root, pathline = [],
                               outlist = [], outpath = None ) :
    if os.path.isdir(root) :
        rootpath, subd, files = next(os.walk(root))
        for f in files :
            outlist += ['.'.join(pathline+f.split('.')[:-1])]
            if outpath is not None :
                outpath += [ os.path.join(rootpath, f) ]
        for s in subd : _recursive_list_ssp_libs(os.path.join(root, s),
                                                 pathline+[s], outlist,
                                                 outpath)
    return;

def print_ssp_libs () :
    """Print on screen a list of the available SSP libraries.
    
    This has the only purpose of listing the possible choices.

    Naming convention: **author.method[.imf.filling_schema]**

    #. **author**: An achronym for the authors of the SSP library
    #. **method**: if present, shows which method was used to compute the SSP
    #. **imf**: if present shows the initial mass function used to compute the SSP
    #. **filling_schema**: all the SSPs' wavelength domain has been extended to
       span from :math:`1\ \mathring{A}` to :math:`10^{10}\ \mathring{A}`. 
       This code provides the strategy used (plain = filled with zeros, 
       :code:`extrap` = extrapolated linearly in the logarithm, 
       :code:`refined` = thinner lambda grid obtained by 
       linearly-interpolating the plain equivalent)
    """
    from galapy.configuration import rcParams
    outlist = []
    for path in rcParams['datapath'] :
        _recursive_list_ssp_libs( root = os.path.join( path, *GP_GBL.SSP_DIR ),
                                  pathline = [], outlist = outlist )
    list_libs = ''
    for k in outlist :
        list_libs += f'* {k};\n'
    print( 'Available SSP formats\n'
           '---------------------\n'
           + list_libs )

def list_ssp_libs ( return_path = False ) :
    """Return a list of the available SSP libraries.
    
    Parameters
    ----------
    return_path : bool
        (default = False) if True, also return a list
        with absolute paths to each of the available SSP libs.
    
    Returns
    -------
    : list
        available SSP libraries
    : list
        available SPP libraries as a list of absolute paths.
    """
    
    from galapy.configuration import rcParams
    outlist = []
    outpath = None
    if return_path :
        outpath = []
    for path in rcParams['datapath'] :
        _recursive_list_ssp_libs( root = os.path.join( path, *GP_GBL.SSP_DIR ),
                                  pathline = [], outlist = outlist,
                                  outpath = outpath )
    if outpath is not None :
        return outlist, outpath
    return outlist

_SSP_LIB = dict( zip( *list_ssp_libs(True) ) )

def load_SSP_table ( which ) :
    """Load a SSP library corresponding to a given name.
    """

    filepath, filename = os.path.split( _SSP_LIB[ which ] )
    path_line = []
    level = None
    while level != GP_GBL.DATA_DIR :
        filepath, level = os.path.split( filepath )
        path_line = [ level ] + path_line
    
    return _loadSSP( DataFile( filename, tuple(path_line) ).get_file() )

def format_SSP_table ( L, lidx, tidx, Zidx, lsize, tsize, Zsize, flat = True ) :
    """Given an input SSP table returns a version with the formatting internally
    required by the library.
        
    Parameters
    ----------
    L : array-like
        input SSP table. It can be either a continuous array indexed on 3 coordinates
        (i.e. wavelength, age and metallicity) in whatever ordering, or a 3-dimensional
        matrix for which the 3 dimensions run over the wavelength, age and metallicity 
        (also in this case the order is not important).
    lidx : integer
        axis number in the array `L` corresponding to the wavelength dimension 
    tidx : integer
        axis number in the array `L` corresponding to the age dimension 
    Zidx : integer
        axis number in the array `L` corresponding to the metallicity dimension 
    lsize : integer
        dimension of the wavelength grid 
    tsize : integer
        dimension of the age grid 
    Zsize : integer
        dimension of the metallicity grid
    flat : bool
        if True, the input array is assumed to be flattened, if False it should instead
        be a 3D matrix with axes corresponding to the 3 grids in wavelength, age and metallicity
        
    Returns
    -------
    newL : 1d contiguous array
        output SSP table in the correct format
    """
    input_shape = numpy.empty(3, dtype=int)
    input_shape[lidx] = lsize
    input_shape[tidx] = tsize
    input_shape[Zidx] = Zsize
    if flat :
        L = numpy.array(L).reshape(input_shape)
    if tuple(input_shape) != L.shape :
        raise RuntimeError( f'Wrong input shape, expected {tuple(input_shape)}, got {L.shape}' )
        
    return numpy.ascontiguousarray( 
        numpy.transpose(L, axes=(Zidx, lidx, tidx)).ravel()
    )

def store_SSP_table ( outfile, l, t, Z, L, endianism = 'little', force = False ) :
    """Save a new SSP table
    
    Parameters
    ----------
    outfile : string
        path to output file name.
    l : array-like
        wavelength grid
    t : array-like
        age grid
    Z : array-like
        metallicity grid
    L : array-like
        flattened SSP table in the correct format accepted by the library
        (check function ``format_SSP_table()``)
    endianism : string
        endianism of your machine, you can check this by calling ``import sys; print(sys.byteorder)``
    force : bool
        whether to overwrite an already existing ``outfile``
    """
    from sys import byteorder
    if endianism != byteorder :
        raise RuntimeError('wrong endianism selected')
    if os.path.isfile(outfile) and not force :
        warnings.warn( 
            f'Selected filename {outfile} is an existing file, to overwrite it set parameter `force=True`')
        return;
    if not os.path.isdir(os.path.dirname(outfile)) :
        raise RuntimeError(
            'The selected output file name is pointing to path that does not exist in the filesystem'
        )
    l = numpy.ascontiguousarray(l)
    t = numpy.ascontiguousarray(t)
    Z = numpy.ascontiguousarray(Z)
    L = numpy.ascontiguousarray(L)
    with open( outfile, 'wb' ) as f :
        # write wavelength block
        f.write( l.size.to_bytes(8, endianism) )
        f.write( l.tobytes() )
        # write age block
        f.write( t.size.to_bytes(8, endianism) )
        f.write( t.tobytes() )
        # write metallicity block
        f.write( Z.size.to_bytes(8, endianism) )
        f.write( Z.tobytes() )
        # write luminosity block
        f.write( L.size.to_bytes(8, endianism) )
        f.write( L.tobytes() )
    return;

def reshape_SSP_table ( L, shape ) :
    """Reshapes a flattened SSP table to the original 3-dimensional shape.
    
    The 3D matrix has the dimensions of the 3 grids over which it varies:
    wavelength (l_grid), age (t_grid) and metallicity (Z_grid).
    
    Parameters
    ----------
    L : 1d numpy array
        Input flattened SSP table with length l_grid*t_grid*Z_grid
    shape : tuple
        Tuple of 3 integers with the sizes of
        
        1. wavelength grid 
        2. age grid
        3. metallicity grid
    
        The above ordering is mandatory
    
    Returns
    -------
    : 3-d numpy array
        3D copy of the original array with shape (l_grid, t_grid, Z_grid).
    """
    return numpy.transpose(
        L.reshape(
            (shape[2], shape[0], shape[1])
        ),
        axes = ( 1, 2, 0 )
    )

class CSP () :
    """ The Composite Stellar Population object handles and manipulates SSP libraries.

    It provides convenient functions for reading the SSPs from the provided formatted 
    binary files. 
    SSPs can be extracted for given age/wavelength/metallicity.
    Providing an object of type galapy.StarFormationHistory the self.emission function
    computes the luminosity of the resulting Composite Stellar Population.
    
    Parameters
    -----------
    ssp_lib : string
      which SSP library to load. The default is :code:`parsec22.NT`-
      To see the list of available libraries run 
      :code:`galapy.CompositeStellarPopulation.print_ssp_libs()`
    CCSN : bool
      whether to allow for Core-Collapse Supernova support (default=False).

    See Also
    --------
    :func:`galapy.CompositeStellarPopulation.print_ssp_libs()`
    """

    def __init__ ( self, ssp_lib = 'parsec22.NT', CCSN = False ) :
        
        self.ssp_lib = None
        if ssp_lib in set(_SSP_LIB.keys()) :
            self.ssp_lib = ssp_lib
        else :
            raise ValueError( f'SSP library "{ssp_lib}" not available; '
                              'you can see the list of available libraries by running '
                              'galapy.CompositeStellarPopulation.print_ssp_libs()' )
        
        self.l, self.t, self.Z, self.L = load_SSP_table( self.ssp_lib )
        self.shape = (self.l.size, self.t.size, self.Z.size)

        self.CCSN = CCSN
        self.core = CCSP( self.l, self.t, self.Z, self.L, self.CCSN )
        self.L = reshape_SSP_table( self.L, self.shape )
        self._timetuple = None

        # steal docstrings from C-core:
        self.SSP.__func__.__doc__      = self.core.SSP.__doc__
        self.emission.__func__.__doc__ = self.core.emission.__doc__

    def set_parameters ( self, age, sfh ) :
        """ Function for setting the internal quantities required for combining SSPs into CSPs.

        Parameters
        ----------
        age : float
          The age in years of the CSP.
        sfh : object of type SFH()
          The chosen star formation history. This has to be (or inherit from) 
          the instance of an object of type galapy.StarFormationHistory.SFH().
        """
        from .StarFormationHistory import SFH
        from .SFH_core import CSFH
        
        if isinstance( sfh, CSFH ) :
            self._timetuple = sfh.time_grid( age, self.t, self.Z )
        elif isinstance( sfh, SFH ) :
            self._timetuple = sfh.core.time_grid( age, self.t, self.Z )
        else :
            raise ValueError( 'Argument sfh must be an instance of either '
                              'galapy.StarFormationHistory.SFH or '
                              'galapy.SFH_core.' )
        
        self.core.set_params( *self._timetuple )
        return;
            
    def SSP ( self, il, it, iz ) :
        """ Indexes the SSP library.
        
        Parameters
        ----------
        il : uint or array-like of uints
          index in the wavelength-grid of pre-computed SSPs
        it : uint or array-like of uints
          index in the time-grid of pre-computed SSPs
        iz : uint or array-like of uints
          index in the metallicity-grid of pre-computed SSPs
        
        Returns
        -------
        SSP[il,it,iz] : float or array of floats
          Element (il,it,iz) of the SSP table.
        """
        if il < 0 or it < 0 or iz < 0 :
            raise IndexError( 'Negative index provided' )
        if il >= self.shape[ 0 ] or it >= self.shape[ 1 ] or iz >= self.shape[ 2 ] :
            raise IndexError( 'Luminosity index out of range' )
        return self.core.SSP( il, it, iz )

    def emission ( self, age, sfh, il = None, ftau = None ) :
        """ Computes the CSP emission at given index in the wavelength-grid.

        It approximates the integral:

        .. math::
           
           L_\\lambda^\\text{CSP}(\\tau') = 
           \\int_0^{\\tau'}\\text{d}\\tau F(\\lambda,\\tau)\\cdot 
           L_\\lambda^\\text{SSP}\\bigl[\\tau, Z_\\ast(\\tau'-\\tau)\\bigr]\\psi(\\tau'-\\tau)


        where :math:`\\tau'` is the age of the CSP, :math:`L_\\lambda^\\text{SSP}[\\tau, Z\\ast]`
        is the luminosity of 
        the Simple Stellar Population at given time :math:`\\tau` and at given stellar metallicity
        :math:`Z_\\ast`, :math:`\\psi(\\tau)` is the Star Formation History and
        :math:`F(\\lambda, \\tau)` is aa attenuating function of choice.
        The latter is passed as the keyword argument :code:`ftau` and should be a matrix
        with dimension :math:`(N_\\lambda, N_\\tau)`, where :math:`N_\\lambda` is the
        size of the wavelength grid (chosen through the keyword argument :code:`il`) and 
        :math:`N_\\tau` is the size of the time grid of the SSP (:code:`self.shape[1]`).
        :code:`ftau` defaults to :code:`None`, in which case no attenuation is applied.

        Parameters
        ----------
        age : float
          The age in years of the CSP.
        sfh : object of type SFH()
          The chosen star formation history. This has to be (or inherit from) 
          the instance of an object of type galapy.StarFormationHistory.SFH().
        il : array of int
           array of indexes of the positions in the wavelength-grid
           for which to compute the emission. The default value is :code:`None`, for which 
           the function will return the emission all over the wavelength grid.
        Ftau : array
           array containing a function of time to convolve the integrand with 
           (must have same dimension of the `il` array size times the SSP's time-grid size)

        Returns
        -------
        L_CSP : array or scalar float
           the emission of the galaxy's CSP filtered by the function of time :math:`F(\\tau)`
        """

        self.set_parameters( age, sfh )
        if il is None :
            il = numpy.arange( len(self.l), dtype = numpy.uint64 )
        il = numpy.asarray( il )
        scalar_input = False
        if il.ndim == 0 :
            il = il[None] # makes il 1D
            scalar_input = True
        if ftau is None :
            ftau = numpy.ones( ( self.t.size * len( il ), ),
                               dtype = numpy.float64 )
        ftau = numpy.ascontiguousarray( ftau.ravel() )

        ret = self.core.emission( il, ftau )
        if scalar_input :
            return ret.item()
        return ret

    def RCCSN ( self, age, sfh ) :
        """
        Parameters
        ----------
        age : float
          The age in years of the CSP.
        sfh : object of type SFH()
          The chosen star formation history. This has to be (or inherit from) 
          the instance of an object of type galapy.StarFormationHistory.SFH().
        
        Returns
        -------
        RCCSN : scalar float
           The Core-Collapse Supernova rate at given age for a given SFH.

        Raises
        ------
        Warning
          if the object has been built without CCSN support.
        """

        if self.CCSN :
            self.set_parameters( age, sfh )
            return self.core.RCCSN()
        else :
            warnings.warn( f"This instance of the {type(self).__name__} "
                           "has been built without CCSN support. "
                           "Build with `CCSN = True` " )
            return 0.
        
    



# External includes
import os

def _find_home():
    """Locates and return the home directory (or best approximation) on this
    system.

    Raises
    ------
    OSError
        If the home directory cannot be located - usually means you are running
        GalaPy on some obscure platform that doesn't have standard home
        directories.

    Notes
    -----
    This routine has been adapted from the Astropy Collaboration (http://www.astropy.org)
    """
    try:
        homedir = os.path.expanduser('~')
    except Exception:
        # Linux, Unix, AIX, OS X
        if os.name == 'posix':
            if 'HOME' in os.environ:
                homedir = os.environ['HOME']
            else:
                raise OSError('Could not find unix home directory to search for '
                              'galapy config dir')
        elif os.name == 'nt':  # This is for all modern Windows (NT or after)
            if 'MSYSTEM' in os.environ and os.environ.get('HOME'):
                # Likely using an msys shell; use whatever it is using for its
                # $HOME directory
                homedir = os.environ['HOME']
            # See if there's a local home
            elif 'HOMEDRIVE' in os.environ and 'HOMEPATH' in os.environ:
                homedir = os.path.join(os.environ['HOMEDRIVE'],
                                       os.environ['HOMEPATH'])
            # Maybe a user profile?
            elif 'USERPROFILE' in os.environ:
                homedir = os.path.join(os.environ['USERPROFILE'])
            else:
                try:
                    import winreg as wreg
                    shell_folders = r'Software\Microsoft\Windows\CurrentVersion\Explorer\Shell Folders'  # noqa: E501
                    key = wreg.OpenKey(wreg.HKEY_CURRENT_USER, shell_folders)

                    homedir = wreg.QueryValueEx(key, 'Personal')[0]
                    key.Close()
                except Exception:
                    # As a final possible resort, see if HOME is present
                    if 'HOME' in os.environ:
                        homedir = os.environ['HOME']
                    else:
                        raise OSError('Could not find windows home directory to '
                                      'search for galapy config dir')
        else:
            # for other platforms, try HOME, although it probably isn't there
            if 'HOME' in os.environ:
                homedir = os.environ['HOME']
            else:
                raise OSError('Could not find a home directory to search for '
                              'galapy config dir - are you on an unsupported '
                              'platform?')
    return homedir

def _find_or_create_root_dir() :
    """ Search for the GalaPy rootdir in user's home directory.
    If it cannot find it, one is created in ``<home>/.galapy``.
    """

    rootdir = os.path.join( _find_home(), '.galapy' )

    # checks the path already exists:
    if not os.path.exists( rootdir ) :
        # first, if the path is not present try to make a new directory:
        try :
            os.mkdir( rootdir )
        except OSError :
            if not os.path.isdir( rootdir ) :
                raise
    elif not os.path.isdir( rootdir ) :
        # if the path exists but it is not a directory raise
        raise OSError( f'Intended galapy root directory {rootdir} is actually a file' )

    # return absolute path to root directory
    return os.path.abspath( rootdir )



#############################################################################################
# External imports

import numpy
import sys

#############################################################################################
# Internal imports

from galapy.internal.utils import now_string

#############################################################################################

emcee_default_sampler_kw = { 'moves' : None, 'args' : None, 'kwargs' : None }
emcee_default_sampling_kw = { 'log_prob0' : None, 'rstate0' : None, 'blobs0' : None,
                              'tune' : False, 'skip_initial_state_check' : False,
                              'thin_by' : 1, 'thin' : None, 'store' : True,
                              'progress' : True, 'progress_kwargs' : None }

#############################################################################################

dynesty_default_sampler_kw = { 'bound' : 'multi', 'sample' : 'rwalk',
                               'update_interval' : 0.6, 'walks' : 25,
                               'bootstrap' : 0 }
dynesty_default_sampling_kw = { 'nlive_init' : 1024, 'maxiter_init' : None,
                                'maxcall_init' : None, 'dlogz_init' : 0.02,
                                'nlive_batch' : 1024, 'maxiter_batch' : None,
                                'maxcall_batch' : None,
                                'maxiter' : sys.maxsize, 'maxcall' : sys.maxsize,
                                'maxbatch' : sys.maxsize,
                                'n_effective' : None, 'print_progress' : True,
                                'stop_kwargs' : { 'target_n_effective' : 10000 } }

#############################################################################################

class Sampler () :
    
    def __init__ ( self, loglikelihood, ndim, 
                   sampler = 'dynesty',
                   prior_transform = None,
                   nwalkers = None,
                   pool = None,
                   dynesty_sampler_kw = {}, 
                   emcee_sampler_kw = {} ) :
        """
        """
        
        self.sampler = None
        self.which_sampler = sampler
        
        ######################################################################################
        # Dynamic Nested-Sampling with dynesty
        if self.which_sampler == 'dynesty' :
            if prior_transform is None or not hasattr(prior_transform, '__call__'):
                raise RuntimeError( 'You have to pass a function to `prior_transform` '
                                    'when running with the dynesty sampler.')

            kw = dynesty_default_sampler_kw
            kw.update( dynesty_sampler_kw )
            from dynesty import DynamicNestedSampler
            self.sampler = DynamicNestedSampler( 
                loglikelihood = loglikelihood,
                prior_transform = prior_transform,
                ndim = ndim,
                pool = pool,
                **kw
            )
            
        ######################################################################################
        # MCMC Ensemble sampling with emcee
        if self.which_sampler == 'emcee' :
            if nwalkers is None :
                raise RuntimeError( 'You have to pass an integer number of walkers to `nwalkers` '
                                    'when running with the emcee sampler.')
            kw = emcee_default_sampler_kw
            kw.update( emcee_sampler_kw )
            from emcee import EnsembleSampler
            self.sampler = EnsembleSampler( nwalkers = nwalkers,
                                            ndim = ndim, 
                                            log_prob_fn = loglikelihood, 
                                            pool = pool, 
                                            **kw )
        
        ######################################################################################
        # Maybe in the future we'll add more sampling options ...    
        if self.sampler is None :
            raise AttributeError( f'The sampler chosen "{self.which_sampler}" is not valid.'
                                  'Valid samplers are ["dynesty", "emcee"].' )
            
    def run_sampling ( self, pos = None, nsample = None,
                       dynesty_sampling_kw = {}, emcee_sampling_kw = {} ) :
        """ Run the sampling with the chosen sampler.
        
        Parameters
        ----------
        pos : iterable
        nsample : int
        
        Returns
        -------
        : None
        """
        from time import time
        
        if self.which_sampler == 'dynesty' :
            kw = dynesty_default_sampling_kw
            kw.update( dynesty_sampling_kw )
            tstart = time()
            self.sampler.run_nested( **kw )
            ndur = time() - tstart
            print( f'\nDone dynesty (dynamic) in {ndur} seconds' )
            # Print summary of the results:
            self.sampler.results.summary()
            
        if self.which_sampler == 'emcee' :
            if pos is None :
                raise RuntimeError( '`pos = None` but you should provide an initial position '
                                    'for the walkers when running with emcee.')
            if nsample is None :
                raise RuntimeError( '`nsample = None` but you should provide a number '
                                    'of samples to draw when running with emcee.')
            kw = emcee_default_sampling_kw
            kw.update( emcee_sampling_kw )
            tstart = time()
            self.emcee_state = self.sampler.run_mcmc( pos, nsample, **kw )
            ndur = time() - tstart
            print( f'\nDone emcee in {ndur} seconds' )
            
        return;

    def return_samples_logl_weights ( self ) :

        outtuple = None
        if self.which_sampler == 'dynesty' :
            weights = numpy.exp( self.sampler.results.logwt -
                                 self.sampler.results.logz[-1] )
            outtuple = ( self.sampler.results.samples,
                         self.sampler.results.logl,
                         weights )
        if self.which_sampler == 'emcee' :
            outtuple = ( self.sampler.flatchain,
                         self.sampler.flatlnprobability,
                         numpy.ones_like( self.sampler.flatlnprobability ) )
        if outtuple is None :
            raise RuntimeError(
                "Function must be called after sampling run"
            )
        
        return outtuple

    def save_results ( self, outbase = '',
                       pickle_sampler = False,
                       pickle_raw = True ) :
        """ Store the results of the last sampling run.
        NOTE THAT the type of output depends on the sampler used.
        
        Parameters
        ----------
        outbase : string
          Position in the filesystem where the results will be stored.
          Default to the directory where the command has been called with
          a current date+time as name.
        pickle_sampler : bool
          If set to `True` the sampler will be pickled for future re-use.
        pickle_raw : bool
          If set to `True` the raw results of the sampling run will be pickled.
        
        Returns
        -------
        : None
        """
        import os
        import pickle

                    
        # If no name for the current run has been provided, set it to
        # a string with the current date+time: 'year+month+day+hour+minute'
        if len(outbase) == 0 :
            outbase = now_string()
            
        if self.which_sampler == 'dynesty' :
            if pickle_raw :
                with open( '_'.join( [ outbase, 'dynesty_results.pickle' ] ), 'wb' ) as f :
                    pickle.dump( self.sampler.results, f )
            if pickle_sampler :
                with open( '_'.join( [ outbase, 'dynesty_sampler.pickle' ] ), 'wb' ) as f :
                    pickle.dump( self.sampler, f )

        if self.which_sampler == 'emcee' :
            if pickle_raw :
                # Pickling sampler-state:
                with open( '_'.join( [ outbase, 'emcee_state.pickle' ] ), 'wb' ) as f :
                    pickle.dump( self.emcee_state, f )
                # Storing the flattened chains:
                numpy.save( '_'.join( [ outbase, 'emcee_flatchain' ] ),
                            self.sampler.flatchain )
                # Storing the flattened probabilities:
                numpy.save( '_'.join( [ outbase, 'emcee_flatlnprob' ] ),
                            self.sampler.flatlnprobability )
                # Storing the acceptance-fraction:
                numpy.save( '_'.join( [ outbase, 'emcee_axeptfrac' ] ),
                            self.sampler.acceptance_fraction )
            if pickle_sampler :
                with open( '_'.join( [ outbase, 'emcee_sampler.pickle' ] ), 'wb' ) as f :
                    pickle.dump( self.sampler, f )

        print( f'Sampler results stored in files with prefix: {outbase}' )

        return;

#############################################################################################



"""
Implements galaxy classes wrapping up the interplay between the different components 
contribution to the overall emission.
By instantiating an object of class ``GXY`` or derived it is possible to set-up the
different components, compute global and component-specific emission, compute the 
derived quantities and modify the parameter setting.
"""

# External imports
import numpy
import os
from collections.abc import MutableMapping as MM

# Internal imports
from galapy.StarFormationHistory import SFH
from galapy.CompositeStellarPopulation import CSP
from galapy.InterStellarMedium import ISM
from galapy.ActiveGalacticNucleus import AGN
from galapy.XRayBinaries import XRB
from galapy.NebularFreeFree import NFF
from galapy.PhotometricSystem import PMS
from galapy.Synchrotron import SNSYN
from galapy.Cosmology import CSM
from galapy.InterGalacticMedium import IGM

from galapy.internal.utils import trap_int, find_nearest
from galapy.internal.interp import lin_interp
from galapy.internal.constants import Lsun, sunL, clight, Mpc_to_cm, hP
import galapy.internal.globs as GP_GBL
from galapy.internal.data import DataFile
from galapy.internal.abc import Model

########################################################################################

gxy_params_defaults = {

    ##########
    # Galaxy #
    ##########

    'age'      : ( 'Age of the galaxy',
                   [6., 11.], True, 'age' ),
    'redshift' : ( 'Redshift of the galaxy',
                   [0., 10.], False, 'redshift' ),

    ##########################
    # Star Formation History #
    ##########################
    
    'sfh.tau_quench' : ( 'Age of the abrupt quenching',
                         [6., 11.], True, '\\tau_\\mathrm{quench}' ),

    # In-Situ model
    'sfh.psi_max' : ( 'Normalisation',
                      [0., 4.], True, '\\psi_\\mathrm{max}' ),
    'sfh.tau_star' : ( 'Characteristic timescale',
                       [6., 11.], True, '\\tau_\\star' ),

    # Constant model
    'sfh.psi'   : ( 'Value of the constant SFR',
                    [0., 4.], True, '\\psi_0' ),
    'sfh.Mdust' : ( 'Total dust mass in galaxy at the given age',
                    [6., 14.], True, 'M_\\mathrm{dust}' ),
    'sfh.Zgxy'   : ( 'Metallicity of all phases in galaxy at the given age',
                     [0., 1.], False, 'Z_\\mathrm{gxy}' ),
    
    # Delayed-Exponential model
    'sfh.psi_norm' : ( 'Normalisation',
                       [0., 4.], True, '\\psi_\\mathrm{norm}' ),
    'sfh.k_shape'  : ( 'Shape parameter of the early evolution',
                       [0., 5.], False, '\\kappa' ),
    'sfh.tau_star' : ( 'Characteristic timescale',
                       [6., 11.], True, '\\tau_\\star' ),
    'sfh.Mdust' : ( 'Total dust mass in galaxy at the given age',
                    [6., 14.], True, 'M_\\mathrm{dust}' ),
    'sfh.Zgxy'   : ( 'Metallicity of all phases in galaxy at the given age',
                     [0., 1.], False, 'Z_\\mathrm{gxy}' ),
    
    # Log-Normal model
    'sfh.psi_norm'   : ( 'Normalisation',
                         [0., 4.], True, '\\psi_\\mathrm{norm}' ),
    'sfh.sigma_star' : ( 'Characteristic width',
                         [0., 5.], False, '\\sigma_\\star' ),
    'sfh.tau_star'   : ( 'Peak age',
                         [6., 11.], True, '\\tau_\\star' ),
    'sfh.Mdust' : ( 'Total dust mass in galaxy at the given age',
                    [6., 14.], True, 'M_\\mathrm{dust}' ),
    'sfh.Zgxy'   : ( 'Metallicity of all phases in galaxy at the given age',
                     [0., 1.], False, 'Z_\\mathrm{gxy}' ),

    ########################
    # Inter-Stellar Medium #
    ########################

    'ism.f_MC' : ( 'Fraction of dust in the MC phase',
                   [0., 1.], False, 'f_\\mathrm{MC}' ),

    ##################
    # Molecular Clouds
    
    'ism.norm_MC' : ( 'Normalisation of the MC extinction in the visible band',
                      [-1., 4.], True, '\\mathcal{C}_{V~\\mathrm{MC}}' ),
    'ism.N_MC'    : ( 'Number of MCs in the galaxy',
                      [0., 5.], True, 'N_\\mathrm{MC}' ),
    'ism.R_MC'    : ( 'Average radius of a MC',
                      [0., 5.], True, 'R_\\mathrm{MC}' ),
    'ism.tau_esc' : ( 'Time required by stars to start escaping their MC',
                      [4., 8.], True, '\\tau_\\mathrm{esc}' ),
    'ism.dMClow'  : ( 'Molecular clouds extinction index at wavelength < 100 mum (10^6 Ang)',
                      [0., 5.], False, '\\delta_{\\mathrm{MC}~l}' ),
    'ism.dMCupp'  : ( 'Molecular clouds extinction index at wavelength > 100 mum (10^6 Ang)',
                      [0., 5.], False, '\\delta_{\\mathrm{MC}~u}' ),
    
    ##############
    # Diffuse Dust
    
    'ism.norm_DD' : ( 'Normalisation of the DD extinction in the visible band',
                      [-1., 4.], True, '\\mathcal{C}_{V~\\mathrm{DD}}' ),
    'ism.Rdust'   : ( 'Radius of the diffuse dust region embedding stars and MCs',
                      [0., 5.], True, 'R_\\mathrm{DD}' ),
    'ism.f_PAH'   : ( 'Fraction of the total DD luminosity radiated by PAH',
                      [0., 1.], False, 'f_\\mathrm{PAH}' ),
    'ism.dDDlow'  : ( 'Diffuse dust extinction index at wavelength < 100 mum (10^6 Ang)',
                      [0., 5.], False, '\\delta_{\\mathrm{DD}~l}' ),
    'ism.dDDupp'  : ( 'Diffuse dust extinction index at wavelength > 100 mum (10^6 Ang)',
                      [0., 5.], False, '\\delta_{\\mathrm{DD}~u}' ),

    ###############
    # Synchrotorn #
    ###############

    'syn.alpha_syn'   : ( 'SN synchrotron spectral index',
                          [0., 5.], False, '\\alpha_\\mathrm{syn}' ), 
    'syn.nu_self_syn' : ( 'Self-absorption frequency of the SN synchrotron',
                          [0., 1.], False, '\\nu_\\mathrm{syn}^\\mathrm{self}'),

    #####################
    # Nebular Free-Free #
    #####################

    'nff.Zgas' : ( 'Metallicity of the ionised gas of nebular regions',
                   [0., 1.], False, 'Z_\\mathrm{gas}' ),
    'nff.Zi'   : ( 'Average atomic number of ions', [0., 10.], False, '\\mathcal{Z}^i'),

    ###########################
    # Active Galactic Nucleus #
    ###########################

    'agn.fAGN' : ( 'Fraction with respect to the total dust IR luminosity contributed by the AGN',
                   [-3., 3.], True, 'f_\\mathrm{AGN}' ),
    
    'agn.ct' : ( 'Torus half-aperture angle', 40, None, '\\Theta' ),
    'agn.al' : ( 'Density parameter (exponential part)', 0., None, '\\alpha' ),
    'agn.be' : ( 'Density parameter (power-law part)', -0.5, None, '\\beta' ),
    'agn.ta' : ( 'Optical depth at 9.7 mum', 6., None, '\\tau_{9.7}^\\mathrm{AGN}' ),
    'agn.rm' : ( 'Radial ratio of the torus', 60, None, 'R_\\mathrm{torus}^\\mathrm{AGN}' ),
    'agn.ia' : ( 'Inclination angle', 0.001, None, '\\Psi_\\mathrm{los}^\\mathrm{AGN}' ),

}

########################################################################################

class GXY ( Model ) :
    """Base galaxy class. All other galaxy models derive from this object.
    All the arguments are optional.
    Whether a particular component is included or not (e.g. radio, x-ray) the wavelength 
    grid will always span from 1 to :math:`10^{10}` angstroms.
    Note that by default, the only components that will be built are 

    1. star formation history (SFH): modelling the stellar mass growth     
    2. composite stellar populations (CSP): modelling unattenuated stellar emission
    3. inter-stellar medium (ISM): modelling the absorption/re-radiation due to 
       inter-stellar dust. This is divided in two components: molecular clouds (MC)
       and diffuse dust (DD)

    All the other components are not strictly necessay to generate an SED and, for the
    sake of performances, they should be built only if necessary. 
    
    Parameters
    ----------
    age : float
        (``None`` defaults to ``1.e+6``) age in years of the galaxy at the time it is observed
    redshift : float
        (``None`` defaults to ``0.0``) redshift of the galaxy
    cosmo : str or dict
        (default = ``'Planck18'``) cosmological model of choice, takes the same arguments 
        an object of type galapy.Cosmology.CSM would take. 
        If a string is passed, it should name one of the pre-computed cosmologies available in the
        database. Available cosmologies are 
        ``'WMAP7'``, ``'WMAP9'``, ``'Planck15'``, ``'Planck18'``.
        If a dictionary is passed, the class expects to find 3 key-value couples:
        key = ``'redshift'``, value = an array of redshift values;
        key = ``'luminosity_distance'``, value = an array of luminosity distances corresponding to 
        the redshift values of the first key-value couple;
        key = ``'age'``, value = an array of ages of the Universe corresponding to the redshift 
        values of the first key-value couple.
        All these arrays should have the same length.
    lstep :  scalar int or boolean mask
        (default = ``None``) Reduces the granularity of the wavelength grid.
        If the input is scalar, select one any `lstep` wavelengths.
        If the input is a boolean mask, only select the indexes of the
        array masked with lstep.
    do_Xray : bool
        (default = ``False``) build the components modelling X-Ray emission, 
        i.e. X-Ray binaries (low & high mass) and, if ``do_AGN=True``, the X-Ray
        part of the spectrum from the eventual AGN.
    do_Radio : bool
        (default = ``False``) build the components modelling Radio emission if necessary.
        If the chosen SSP library does not already include nebular and synchrotron emission,
        ``do_Radio = True`` will build objects of type NFF (nebular free-free) and SNSYN 
        (super-nova synchrotron). The eventual parameters of both components can be passed
        through arguments ``nff`` and ``syn``, respectively.
        Note that if ``do_Radio = False`` both ``nff`` and ``syn`` will be ignored.
    do_AGN : bool
        (default = ``False``) build AGN component. The AGN spectrum is modelled through 
        templates from Fritz et al. (2006) in the UV to IR bands and with an attenuated
        power-law in the X-ray band (only if also argument ``do_Xray`` is set to ``True``.
        Tunable parameters can be set by passing them to the argument ``agn``.
    do_IGM : bool
        (default = ``False``) build attenuation due to inter-galactic hydrogen with
        model from Inoue et al., (2014).
    sfh : dict
        (default = ``None``) arguments passed to the SFH object builder, if ``None`` is passed
        uses the default parameter of class ``galapy.StarFormationHistory.SFH``
    csp : dict
        (default = ``None``) arguments passed to the CSP object builder, if ``None`` is passed
        uses the default parameter of class ``galapy.CompositeStellarPopulation.CSP``
    ism : dict
        (default = ``None``) arguments passed to the ISM object builder, if ``None`` is passed
        uses the default parameter of class ``galapy.InterStellarMedium.ISM``
    agn : dict
        (default = ``None``) arguments passed to the AGN object builder, if ``None`` is passed
        uses the default parameter of class ``galapy.ActiveGalacticNucleus.AGN``. 
        If ``do_AGN=False`` this argument is ignored.
    nff : dict
        (default = ``None``) arguments passed to the NFF object builder, if ``None`` is passed
        uses the default parameter of class ``galapy.NebularFreeFree.NFF``. 
        If ``do_Radio=False`` this argument is ignored.
    syn : dict
        (default = ``None``) arguments passed to the SNSYN object builder, if ``None`` is passed
        uses the default parameter of class ``galapy.Synchrotron.SNSYN``. 
        If ``do_Radio=False`` this argument is ignored.
    """
    
    def __init__ ( self, age = None, redshift = None,
                   cosmo = 'Planck18', lstep = None,
                   do_Xray = False, do_Radio = False,
                   do_AGN = False, do_IGM = False,
                   sfh = None, csp = None, ism = None,
                   agn = None, nff = None, syn = None ) :

        # Define a dictionary for parameters storage
        self.params = {}

        # Define a dictionary to store components
        self.components = {
            'stellar' : None,
            'extinct' : None,
            'MC' : None,
            'DD' : None,
        }

        # If no age is passed set it to the minimum
        # possible age allowed by the time-integration grid
        if age is None :
            self.age = 1.e+6
        else :
            self.age = age
        self.params[ 'age' ] = self.age

        # If no redshift is passed set it to zero
        if redshift is None :
            self.redshift = 0.
        else :
            self.redshift = redshift
        self.params[ 'redshift' ] = self.redshift

        # Build the redshift-dependent cosmological quantities:
        self.cosmo = CSM( cosmo )

        # Compute Universe Age at given redshift
        self.UA = self.cosmo.age( self.redshift )

        # Check that age and redshift are compatible
        if self.age > self.UA :
            raise RuntimeError( f"The age of the galaxy (t={self.age:e} years) "
                                f"cannot be larger than the age of the Universe which, "
                                f"at the given redshift ( z = {self.redshift:.3f} ),"
                                f" is {self.UA:e} years." )

        # Instantiate empty dictionaries for the
        # mandatory components when None is passed.
        if sfh is None :
            sfh = {}
        if csp is None :
            csp = {}
        if ism is None :
            ism = {}

        # Build the SFH model and set the parameters of ISM
        # to the values derived from SFH
        self.sfh = SFH( **sfh )
        self.params[ 'sfh' ] = self.sfh.params
        ism.update( { 'Zgas'  : self.sfh.core.Zgas(self.age), 
                      'Mgas'  : self.sfh.core.Mgas(self.age),
                      'Mdust' : self.sfh.core.Mdust(self.age) } )
        
        # tell the CSP constructor to build with
        # SN support if Radio support is requested
        csp[ 'CCSN' ] = do_Radio
        # NOTE THAT a more consistent implementation would be
        # to also set the parameter to false if synchrotron
        # is already included in the SSPs, left for future development:
        # csp[ 'CCSN' ] = do_Radio and 'parsec22.NT' not in csp['ssp_lib']

        self.csp = CSP( **csp )
        self.csp.set_parameters( self.age, self.sfh )

        self.ism = ISM( **ism )
        self.params[ 'ism' ] = {}
        self.params[ 'ism' ].update( self.ism.mc.params )
        self.params[ 'ism' ].update( self.ism.dd.params )
        
        self.xrb = None
        if do_Xray :
            self.xrb = XRB( lmin = self.csp.l.min(),
                            lmax = self.csp.l.max(),
                            age = self.age,
                            psi = self.sfh( self.age ),
                            Mstar = self.sfh.Mstar( self.age ),
                            Zstar = self.sfh.Zstar( self.age ) )
            self.components['XRB'] = None
            # This here below not really necessary as there is
            # no freedom in setting the X-ray Binaries parameters:
            # self.params[ 'xrb' ] = self.xrb.params
            
        self.agn = None
        if do_AGN :
            if agn is None :
                agn = {}
            self.agn = AGN( lmin = self.csp.l.min(),
                            lmax = self.csp.l.max(),
                            do_Xray = do_Xray,
                            **agn )
            self.params[ 'agn' ] = self.agn.params
            self.components['AGN'] = None

        self.nff   = None
        self.snsyn = None
        # Build the Nebular-Free support only if
        # it is not already included in the SSP library
        if do_Radio and 'parsec22.NTL' not in self.csp.ssp_lib :
            if nff is None :
                nff = {}
            self.nff = NFF( self.csp.l, **nff )
            self.Q_H_fact = 1. / clight['A/s'] / hP['erg*s']
            self.w912 = self.csp.l <= 912.
            self.params[ 'nff' ] = self.nff.params
            self.components['nebular'] = None

            # Build the Synchrotron support only if
            # it is not already included in the SSP library
            if 'parsec22.NT' not in self.csp.ssp_lib :
                if syn is None :
                    syn = {}
                syn[ 'RCCSN' ] = self.csp.core.RCCSN()
                self.snsyn = SNSYN( self.csp.l, **syn )
                self.params[ 'syn' ] = self.snsyn.params
                self.components['synchrotron'] = None

        # If sub-gridding is required, build subgrid
        self.lstep = lstep
        if self.lstep is not None :
            self.lgrid = self.get_wavelength_grid(self.lstep)
        else : 
            self.lgrid = numpy.arange(self.csp.l.size)

        # If IGM transmission is required, build it
        # and compute transmission.
        self.igm = None
        self.igm_trans = numpy.ones_like(
            self.wl( obs = True )
        )
        if do_IGM :
            self.igm = IGM()
            self.igm_trans = self.igm.transmission(
                self.wl( obs = True ),
                self.redshift
            )

    def dump ( self ) :

        return dict(
            lstep    = self.lstep,
            do_Xray  = self.xrb is not None,
            do_Radio = self.nff is not None,
            do_AGN   = self.agn is not None,
            do_IGM   = self.igm is not None,
            cosmo    = { 'redshift' : self.cosmo.DL.get_x(),
                         'luminosity_distance' :  self.cosmo.DL.get_y(),
                         'age' :  self.cosmo.age.get_y() },
            csp = { 'ssp_lib' : self.csp.ssp_lib },
            **self.params
        )

    @classmethod
    def load ( cls, dictionary ) :
        return cls( **dictionary )
        
    def wl ( self, obs = False ) :
        """Returns the wavelength grid with the mask applied.
        
        Parameters
        ----------
        obs : bool
            (Optional, default = ``False``) if set to ``True``
            returns the observer's frame wavelength grid, otherwise
            the rest-frame grid is returned.
        
        Returns
        -------
        : 1d-array
        wavelength grid in observer's frame (``obs = True``) or in 
        rest frame (``obs = False``)
        """
        if obs :
            return ( 1 + self.redshift ) * self.csp.l[ self.lgrid ]
        return self.csp.l[self.lgrid]
    
    def get_wavelength_grid ( self, lstep ) :
        """ Reduces the granularity of the wavelength grid [optimization]
        
        Parameters
        ----------
        lstep : scalar int or boolean mask
            If the input is scalar, select one any `lstep` wavelengths.
            If the input is a boolean mask, only select the indexes of the
            array masked with lstep.
            
        Returns
        -------
        : integer array
            A list of indices of the wavelength grid
        """
        if isinstance(lstep, int) :
            return numpy.arange(self.csp.l.size, step=lstep)
        try :
            return numpy.arange(self.csp.l.size)[lstep]
        except IndexError :
            raise TypeError('Argument lstep should be either an integer or a boolean mask!')
        
    def set_parameters ( self, age = None, redshift = None,
                         sfh = None, ism = None,
                         agn = None, nff = None,
                         syn = None ) :
        """Preferred method for changing the value of free-parameters. 
        
        Parameters
        ----------
        age : float
            age of the galaxy
        redshift : float
            redshift of the galaxy
        sfh : dict
            (default = ``None``) arguments passed to SFH.set_parameters method, 
            if ``None`` is passed maintains the current parameterisation 
        csp : dict
            (default = ``None``) arguments passed to the CSP.set_parameters method, 
            if ``None`` is passed maintains the current parameterisation 
        ism : dict
            (default = ``None``) arguments passed to the ISM.set_parameters method, 
            if ``None`` is passed maintains the current parameterisation 
        agn : dict
            (default = ``None``) arguments passed to the AGN.set_parameters method, 
            if ``None`` is passed maintains the current parameterisation  
            If the GXY object has been built without AGN component this argument is ignored.
        nff : dict
            (default = ``None``) arguments passed to the NFF.set_parameters method, 
            if ``None`` is passed maintains the current parameterisation  
            If the GXY object has been built without AGN component this argument is ignored.
        syn : dict
            (default = ``None``) arguments passed to the SNSYN.set_parameters method, 
            if ``None`` is passed maintains the current parameterisation  
            If the GXY object has been built without AGN component this argument is ignored.

        Note
        ----
        The method is built to optimise the number of computations performed. 
        Do not pass arguments that would not change the current parameterisation.
        **e.g.1** passing ``sfh = {}`` is less performant than sticking to the 
        default ``sfh = None``.
        **e.g.2** passing the same value at each call is a waste of computational time: 
        ``redshift = 2.0`` passed at each call will considerably slow down execution.
        **Bottom line**: pass an argument only when necessary. 
        """

        # if age of sfh change, reset the csp's timetuple
        reset_csp = False
        reset_ism = False
        
        if redshift is not None :
            self.redshift = redshift
            self._z = 1. / ( 1 + self.redshift )
            self.UA = self.cosmo.age( self.redshift )
            self.params['redshift'] = self.redshift
            if self.igm is not None :
                self.igm_trans = self.igm.transmission(
                    self.wl( obs = True ),
                    self.redshift
                )
            
        if age is not None :
            # if the provided age is larger than the age of the Universe,
            # set the age to the age of the Universe
            self.age = age
            reset_csp = True
            reset_ism = True
            self.params['age'] = self.age

        if self.age > self.UA :
            raise RuntimeError( "Trying to set an age larger than the age "
                                "of the Universe at current redshift." )

        if sfh is not None :
            self.sfh.set_parameters(**sfh)
            reset_csp = True
            reset_ism = True
            self.params['sfh'].update(self.sfh.params)

        if reset_csp :
            self.csp.set_parameters( self.age, self.sfh )
            if self.snsyn is not None :
                if syn is None :
                    syn = {}
                syn['RCCSN'] = self.csp.core.RCCSN()

        ############################################################################
        # ISM:
        # --------------
        # None  | Reset 
        # ------+-------
        # True  | True
        # True  | False
        # False | True
        # False | False
        # ______|_______
        # --------------
        # This should cover all the possibilities:
        if ism is not None :
            if reset_ism :
                ism.update( { 'Zgas'  : self.sfh.core.Zgas(self.age), 
                              'Mgas'  : self.sfh.core.Mgas(self.age),
                              'Mdust' : self.sfh.core.Mdust(self.age) } )
            self.ism.set_parameters(**ism)
            self.params['ism'].update(self.ism.mc.params)
            self.params['ism'].update(self.ism.dd.params)
        else :
            if reset_ism :
                ism = { 'Zgas'  : self.sfh.core.Zgas(self.age), 
                        'Mgas'  : self.sfh.core.Mgas(self.age),
                        'Mdust' : self.sfh.core.Mdust(self.age) }
                self.ism.set_parameters(**ism)
                self.params['ism'].update(self.ism.mc.params)
                self.params['ism'].update(self.ism.dd.params)

        ############################################################################
        # This one here should also be set only when either age or sfh changes:
        if self.xrb is not None :
            self.xrb.set_parameters( age = self.age,
                                     psi = self.sfh( self.age ),
                                     Mstar = self.sfh.Mstar( self.age ),
                                     Zstar = self.sfh.Zstar( self.age ) )

        ############################################################################
        # AGN varies only if some new parameters are passed
        if agn is not None :
            try :
                self.agn.set_parameters(**agn)
                self.params['agn'].update(self.agn.params)
            except AttributeError :
                raise AttributeError( 'Passing AGN-parameters to a GXY-class built '
                                      'without an AGN component is not allowed.' )
            except :
                raise

        ############################################################################
        # NFF varies both whether some new parameters are passed or if some
        # of the other components above which they do depend changes
        if self.nff is not None :
            if nff is None :
                nff = {}
            nff.update( { 'Zgas' : self.sfh.core.Zgas(self.age) } )
            self.nff.set_parameters( **nff )
            self.params['nff'].update(self.nff.params)

        ############################################################################
        # If the object exists and CCSN varied for some reason, csp has already
        # taken care of modifying the syn-dictionary, otherwise this enters only
        # when some new parameter has been set.
        if syn is not None:
            try :
                self.snsyn.set_parameters( **syn )
                self.params['syn'].update(self.snsyn.params)
            except AttributeError :
                raise AttributeError( 'Passing SYN-parameters to a GXY-class built '
                                      'without an SNSYN component is not allowed.' )
            
        return;
    
    def Lstellar ( self ) :
        """Unattenuated stellar emission in solar luminosities. 
        Approximates the integral
        
        .. math::
           
           L_\\lambda^\\text{CSP}(\\tau') = 
           \\int_0^{\\tau'}\\text{d}\\tau 
           L_\\lambda^\\text{SSP}\\bigl[\\tau, Z_\\ast(\\tau'-\\tau)\\bigr]\\psi(\\tau'-\\tau)

        where :math:`\\tau'` is the age of the galaxy, 
        :math:`L_\\lambda^\\text{SSP}[\\tau, Z\\ast]`
        is the luminosity of the Simple Stellar Population at given time :math:`\\tau` 
        and at given stellar metallicity :math:`Z_\\ast`, :math:`\\psi(\\tau)` is the 
        Star Formation History at time :math:`\\tau`
        
        Returns
        -------
        : 1-d array
        stellar emission on the default wavelength grid
        """
        
        # emission from stars (using directly the core function for performance)  
        return self.csp.core.emission( self.lgrid )

    def _passive_emission ( self, ftau = None ) :
        import warnings
        warnings.filterwarnings("error")
        if ftau is None :
            ftau = numpy.ones( self.csp.shape[:-1] )
        else :
            ftau = ftau.reshape( self.csp.shape[:-1] )
        
        itmin = find_nearest(self.csp.t, self.age-self.sfh.params['tau_quench']).item()
        itmax = max(find_nearest(self.csp.t, self.age).item(), itmin+1)
        itque = find_nearest(self.csp.t, self.sfh.params['tau_quench']).item()
    
        Mweight = trap_int(
            self.csp.t[:itque,numpy.newaxis], 
            (self.sfh( self.csp.t[:itque] ) * ftau[:,:itque]).T
        )
        if itmax-itmin > 1 :
            Lavg = self.csp.core.emission(self.lgrid, ftau) #.astype('float64')
            LocWeight = trap_int( 
                self.csp.t[itmin:itmax, numpy.newaxis], 
                (self.sfh( self.csp.t[itmin:itmax] ) * ftau[:,itmin:itmax]).T 
            )
            # Note that the '>1.e-20' is a NOT ELEGANT SOLUTION for the overflow bug
            # previously present in the following line
            # Lavg[wLW] /= LocWeight[wLW] -> RuntimeWarning: overflow encountered in divide
            # It is generated by Lavg being 'float32' and LocWeight 'float64' but 
            # if Lavg is casted to 'float32' another overflow arises in the loglikelihood computation
            # so this is not a perfect solution and should be investigated.
            wLW = (LocWeight > 1.e-20)
            Lavg[wLW] /= LocWeight[wLW]
            return Lavg * Mweight 
        return self.csp.L[:, itmax, self.csp._timetuple[2][0]] * Mweight
    
    def get_emission ( self, store_attenuation = False, **kwargs ) :
        """Computes the overall emission coming from a galaxy with given parameterisation.
        The resulting shape of the SED depends on the components the galaxy object has
        been built with. This authomatically deals with the interplay among the different 
        active components.

        Parameters
        ----------
        store_attenuation : bool
            (Optional, default = ``False``) if set to ``True``, stores the total,
            wavelength dependent, attenuation due to ISM in an internal variable (``Aavg``)

        **kwargs : dictionary, optional
            arguments passed to function ``set_parameters()``
        
        Returns
        -------
        : 1d-array
            the emission on the selected wavelength grid 
            in units of solar luminosities (:math:`[L_\odot]`)
        
        Note
        ----
        Even though only the overall emission is returned, the contribution of each
        component is stored in the internal dictionary attribute ``components``
        """
        if len( kwargs ) > 0 :
            self.set_parameters( **kwargs )
        
        # attenuation from ISM
        attTotMC, attTot = self.ism.total_attenuation( self.wl(), self.csp.t )
        
        # emission from stars (using directly the core function for performance)
        Lunatt, LattMC, Ltot = self.csp.core._kernel_emission( self.lgrid, attTotMC, attTot )
        # if self.params['age'] <= self.params['sfh']['tau_quench'] :
        #     Lunatt, LattMC, Ltot = self.csp.core._kernel_emission( self.lgrid, attTotMC, attTot )
        # else :
        #     Lunatt = self._passive_emission()
        #     LattMC = self._passive_emission( attTotMC )
        #     Ltot   = self._passive_emission( attTot )
        self.components['stellar'] = Lunatt
        self.components['extinct'] = numpy.array(Ltot) # copy the value

        if store_attenuation or self.nff is not None :
            wn0 = Lunatt > 0.
            self.Aavg = numpy.ones_like( self.wl() )
            self.Aavg[wn0] = Ltot[wn0]/Lunatt[wn0]
        
        # set the resulting temperature of ISM
        EDD = Lsun * trap_int( 
            self.wl(), ( LattMC - Ltot )
        ) 
        EMC = Lsun * trap_int( 
            self.wl(), ( Lunatt - LattMC )
        )
        _ = self.ism.dd.temperature( EDD )
        _ = self.ism.mc.temperature( EMC )
         
        # emission from ISM
        self.components['MC'] = self.ism.mc.emission( self.wl() )
        self.components['DD'] = self.ism.dd.emission( self.wl() )
        Ltot += self.components['MC'] + self.components['DD']

        # OPTIONAL COMPONENTS:

        # emission from AGN
        if self.agn is not None :
            self.components['AGN'] = self.agn.emission( self.wl(), (EDD+EMC) * sunL )
            Ltot += self.components['AGN']

        # emission from X-Ray binaries
        if self.xrb is not None :
            self.components['XRB'] = self.xrb.emission( self.wl() )
            Ltot += self.components['XRB']

        # Free-free emission from Nebulae
        if self.nff is not None :
            Q_H = trap_int( self.wl()[ self.w912 ],
                            Lunatt[ self.w912[ self.lgrid ] ] *
                            self.wl()[ self.w912 ] * self.Q_H_fact )
            self.components['nebular'] = self.Aavg * self.nff.emission( Q_H, il = self.lgrid )
            Ltot += self.components['nebular']

        # Synchrotron emission from SN-shocks
        if self.snsyn is not None :
            self.components['synchrotron'] = self.snsyn.emission( self.lgrid ) * sunL
            Ltot += self.components['synchrotron']
        
        return Ltot

    def get_avgAtt ( self ) :
        """Returns the average attenuation in absolute magnitudes 
        """
        if self.components['stellar'] is None or self.components['extinct'] is None :
            raise RuntimeError(
                "Cannot compute average attenuation if the components dictionary "
                "has not been computed. First run function GXY.get_emission()"
            )
        wn0 = self.components['stellar'] > 0
        AA = numpy.zeros_like( self.wl() )
        AA[wn0] = -2.5 * numpy.log10(
            self.components[ 'extinct' ][ wn0 ] / self.components[ 'stellar' ][ wn0 ]
        )
        return AA

    def get_SED ( self ) :
        """Returns the flux at given distance in units of milli-Jansky [mJy].

        .. math::
            
            S(\\lambda_O) = \\lambda_R^2 \cdot 
            \\dfrac{L_\\text{tot}(\\lambda_R) (1 + z)}{4\\;\\pi\\;c\\;D_L^2(z)} \cdot
            e^{-\\tau_\\text{IGM}(z)}

        where :math:`\\lambda_R` and :math:`\\lambda_O` are the rest-frame and observer's frame
        wavelength, respectively, :math:`L_\\text{tot}` is the total luminosity (as returned by
        function ``get_emission()``), :math:`z` is the redshift, :math:`c` the light-speed,
        :math:`D_L(z)` is the luminosity distance at observed redshift and, if it is included in 
        the model, :math:`e^{-\\tau_\\text{IGM}(z)}` is the IGM transmission at observed redshift.
        
        Returns
        -------
        : 1d-array
        SED flux in milli-Jansky
        """
        return self.cosmo.to_flux(
            self.redshift, self.wl(), self.get_emission()
        ) * self.igm_trans

    def components_to_flux ( self ) :
        """Utility function converting emissions in the internal ``components`` 
        dictionary to fluxes.
        
        Returns
        -------
        : dict
        copy of the internal dictionary ``components`` with the emission conveted 
        to fluxes in milli-Jansky.
        """
        return {
            k : self.cosmo.to_flux( self.redshift, self.wl(), c ) 
            for k, c in self.components.items()
        }
    

class PhotoGXY ( GXY ) :
    """Galaxy class including photometric system.
    This is a class derived from ``galapy.Galaxy`` which implements authomatic computation 
    of fluxes convolved with bandpass transmission filters.
    
    Parameters
    ----------
    pms : PMS instance, optional
      an instance of type ``galapy.PhotometricSystem.PMS``, if not passed,
      it should be built using function ``build_photometric_system``
    *args : tuple, optional
      arguments to be passed to the constructor of the base class ``GXY``
    **kwargs : dictionary, optional
      keyword arguments to be passed to the constructor of the base class ``GXY``
    
    See Also
    --------
    galapy.PhotometricSystem.PMS : class implementing the photometric system
    build_photometric_system : method for building a photometric system internally
    GXY : base class
    """

    def __init__ ( self, *args, pms = None, **kwargs ) :

        super().__init__( *args, **kwargs )
        self.pms = pms

    def dump ( self ) :
        return dict(
            pms_kwargs = self.pms.dump(),
            **super().dump()
        )

    @classmethod
    def load ( cls, dictionary ) :
        # ensure deep-copy
        dictionary = dict( dictionary )
        pms = PMS.load( dictionary.pop('pms_kwargs') )
        return cls( pms = pms, **dictionary )
    
    def build_photometric_system ( self, *args, **kwargs ) :
        """Forwards ``args`` and ``kwargs`` to the constructor of the photometric system.
        
        Parameters
        ----------
        *args : sequence
            positional arguments of the PMS contructor
        **kwargs : dictionary
            keyword arguments of the PMS constructor
        
        See Also
        --------
        galapy.PhotometricSystem.PMS : constructor of the photometric system class
        """

        self.pms = PMS( *args, **kwargs)
        return;

    def photoSED ( self ) :
        """Computes and returns the photometric bandpass fluxes in milliJansky.
        """

        if self.pms is None :
            raise Exception( "Photometric system has not been set. "
                             "Call function build_photometric_system() before." )
        return numpy.asarray(
            self.pms.get_fluxes( self.wl( obs = True ), self.get_SED() )
        )
        

class SpectralGXY ( GXY ) :
    """Not implemented yet
    """
    pass




""" The Cosmology module defines a Cosmology class whose just a wrapper for interpolation on grid of Luminosity Distances and Ages
"""

# External imports
import numpy
from collections.abc import MutableMapping as MM

# Internal imports
import galapy.internal.globs as GP_GBL
from galapy.internal.constants import Lsun, clight, Mpc_to_cm
from galapy.internal.interp import lin_interp
from galapy.internal.data import DataFile


class CSM () :
    """ Class for cosmological computations. It is built using pre-computed redshift-dependent
    quantities useful for transforming energies radiated per unit wavelength into fluxes from
    objects at a given distance.
    
    Parameters
    ----------
    cosmo : string or dictionary
      If a string is passed, it should name one of the pre-computed cosmologies available in the
      database. Available cosmologies are :code:`('WMAP7', 'WMAP9', 'Planck15', 'Planck18')`.
      If a dictionary is passed, the class expects to find 3 key-value couples:
      (i) key = 'redshift', value = an array of redshift values; (ii) key = 'luminosity_distance', 
      value = an array of luminosity distances corresponding to the redshift values of the first key-value couple;
      (iii) key = 'age', value = an array of ages of the Universe corresponding to the redshift 
      values of the first key-value couple.
      All these arrays should have the same length.
    
    Attributes
    ----------
    At construction the class builds 2 interpolator-objects: :code:`CSM.DL` and :code:`CSM.age`.
    These objects provide an interpolation interface to the pre-computed values of luminosity 
    distance and age of the Universe.
    They can be called as functions:
    
    .. code ::
      
      $ csm = CSM('Planck18')
      $ csm.DL(1.0)
      6791.26894

    The returned value is in MegaParsecs for the luminosity distance and in years for the age.
    """
    
    def __init__ ( self, cosmo ) :
        
        if isinstance( cosmo, str ) :
            zz, DL, UA = numpy.loadtxt( DataFile( f'{cosmo:s}.txt',
                                                  GP_GBL.COSMO_DIR ).get_file(),
                                        unpack = True )
        elif isinstance( cosmo, MM ) :
            zz, DL, UA = cosmo['redshift'], cosmo['luminosity_distance'], cosmo['age']
        else :
            raise ValueError( 'Argument `cosmo` should be either a string '
                              'or a formatted dictionary.' )
        zz = numpy.ascontiguousarray(zz)
        DL = numpy.ascontiguousarray(DL)
        TF = numpy.ascontiguousarray(
            1.e+26 * (1 + zz) * Lsun /
            ( 4 * numpy.pi * clight['A/s'] * DL**2 * Mpc_to_cm**2 )
        )
        UA = numpy.ascontiguousarray(UA)
        self.DL  = lin_interp( zz, DL )
        self.age = lin_interp( zz, UA )
        self._TF = lin_interp( zz, TF )

    def to_flux ( self, redshift, restframe_wavelength, luminosity )  :
        """ Converts a restframe luminosity to the flux received at 
        a given redshift.
        
        .. math::
        
          S_{\lambda_O} = \lambda_R^2 L_\text{tot}(\lambda_R) (1+z) / (4 \pi c D_L^2 )
        
        Parameters
        ----------
        redshift : float
        restframe_wavelength : array-like 
        luminosity : array-like
        
        Returns
        -------
        : array-like
          a flux in milliJansky
        """
        return luminosity * restframe_wavelength**2 * self._TF( redshift ) 




import os, h5py, numpy, warnings
from collections.abc import MutableMapping as MM

def recursive_create ( group, indict ) :
    """Dumps an input dictionary to a HDF5 file
    """
    for k, v in indict.items() :
        try :
            if isinstance( v, MM ) :
                new_group = group.create_group( k )
                recursive_create( new_group, v )
            else :
                if hasattr( v, '__len__' ) and not isinstance( v, str ) :
                    group.create_dataset( k, data = v )
                else :
                    if v is not None :
                        group.attrs[k] = v
                    else :
                        group.attrs[k] = 'none'
        except Exception as err :
            print( f"- Skipping field '{k}' as it raises "
                   f"exception {type(err).__name__} with message:\n  '{err}'" )
    return;

def recursive_read ( group, outdict ) :
    """Reads a nested dictionary from a HDF5 file
    """
    outdict.update( { k : v if v != 'none' else None for k,v in dict(group.attrs).items()} )
    for k, v in group.items() :
        try :
            if isinstance( v, h5py.Dataset ) :
                outdict[k] = group.get(k)[:]
            if isinstance( v, h5py.Group ) :
                outdict[k] = {}
                recursive_read( v, outdict[k] )
        except Exception as err :
            print( f"- Skipping dataset '{k}' as it raises "
                   f"exception {type(err).__name__} with message:\n  '{err}'" )
            
    return;

def write_to_hdf5 ( outfile, metadata, hard = False, **groups ) :
    """Automatic dump of several dictionaries into hdf5 file as separate groups in file.
    Also support generic metadata
    
    Parameters
    ----------
    outfile : string
        /path/to/outputfile.hdf5 destination file. The path should exist on the filesystem
    metadata : dict
        dictionary with global metadata
    hard : bool
        (Optional, default=False) if True eventually overwrites an already existing
        file with the same name
    **groups : Keyword arguments
        (Optional) Dictionaries to save to the file. Each dictionary will be saved to the
        output file as a group with its corresponding keyword name as keyword.

    Returns
    -------
    None
    
    See Also
    --------
    load_from_hdf5 : loads all the groups in a HDF5 file
    """

    # check for overwrite
    if not hard and os.path.isfile( outfile ) :
        raise IOError( f"file\n{outfile}\nalready exist, "
                              "to overwrite it set argument hard=True." )

    # check directory exists
    outdir = os.path.dirname(outfile)
    if len( outdir ) == 0 :
        outdir = os.path.curdir
        outfile = os.path.join(outdir,outfile)
    elif not os.path.isdir( outdir ) :
        raise IOError( 'the provided path {os.path.dirname(outfile)} does not exist' )

    with h5py.File( outfile, "w" ) as f :
            
        # store meta-data
        for k,v in metadata.items() :
            f.attrs[k] = v

        # store groups:
        for name, dictionary in groups.items() :
            group = f.create_group(name)
            recursive_create( group, dictionary )
        
    print( f"Wrote on file {outfile}" )
    return;

def load_from_hdf5 ( infile ) :
    """Loads the groups in a hdf5 file into a dictionary.

    Parameters
    ----------
    infile : string
        /path/to/inputfile.hdf5 input file
    
    Returns
    -------
    : dict
        nested dictionary 
    
    See Also
    --------
    write_to_hdf5 : dump of several dictionaries into hdf5 file as separate groups in file
    
    Warning
    -------
    This is a simplified reader intended to work with hdf5 files generate
    using the ``galapy.io.hdf5.write_to_hdf5`` method.
    It is not guaranteed to work with any HDF5. 
    It assumes that at the top level all the fields can be cast to dictionaries.
    If this is not possible it raises a warning.
    """

    # check if file exists
    if not os.path.isfile( infile ) :
        raise IOError( f"file {infile} does not exist." )

    outdict = {}
    with h5py.File( infile, "r" ) as f :
        
        # load meta-data
        outdict['metadata'] = dict(f.attrs)

        # load groups
        for name, todict in f.items() :
            outdict[name] = {}
            recursive_read( todict, outdict[name] )
            
    print( f"Read from file {infile}" )
    return outdict
